""" Module containing custom exception classes
"""


class CostException(Exception):
    """Exception raised by molecule cost classes"""


class ExternalModelAPIError(Exception):
    """Custom error type to signal failure in External model"""


class MoleculeException(Exception):
    """An exception that is raised by molecule class"""


class NodeUnexpectedBehaviourException(Exception):
    """Exception that is raised if the tree search is behaving unexpectedly."""


class PolicyException(Exception):
    """An exception raised by policy classes"""


class RejectionException(Exception):
    """An exception raised if a retro action should be rejected"""


class ScorerException(Exception):
    """Exception raised by scoring classes"""


class StockException(Exception):
    """An exception raised by stock classes"""


class TreeAnalysisException(Exception):
    """Exception raised when analysing trees"""



""" Module containing classes and routines for the GUI interface
"""
import argparse
import logging
import os
import signal
import subprocess
import tempfile
from typing import TYPE_CHECKING

import ipywidgets as widgets
import jupytext
from IPython.display import HTML, display
from ipywidgets import (
    BoundedIntText,
    Button,
    Checkbox,
    Dropdown,
    HBox,
    IntText,
    Label,
    Output,
    SelectMultiple,
    Text,
    VBox,
)
from rdkit import Chem

from aizynthfinder.aizynthfinder import AiZynthFinder
from aizynthfinder.interfaces.gui.utils import pareto_fronts_plot, route_display
from aizynthfinder.utils.logging import setup_logger

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import StrDict


class AiZynthApp:
    """
    Interface class to be used in a Jupyter Notebook.
    Provides a basic GUI to setup and analyze the tree search.

    Should be instantiated with the path of a yaml file with configuration:

    .. code-block::

        from aizynthfinder.interfaces import AiZynthApp
        configfile = "/path/to/configfile.yaml"
        app = AiZynthApp(configfile)

    :ivar finder: the finder instance

    :param configfile: the path to yaml file with configuration
    :param setup: if True will create and display the GUI on instantiation, defaults to True
    """

    def __init__(self, configfile: str, setup: bool = True) -> None:
        setup_logger(logging.INFO)
        self.finder = AiZynthFinder(configfile=configfile)
        self._input: StrDict = dict()
        self._output: StrDict = dict()
        self._buttons: StrDict = dict()
        if setup:
            self.setup()

    def setup(self) -> None:
        """
        Create the widgets and display the GUI.
        This is typically done on instantiation, but this method
        is for more advanced uses.
        """
        self._create_input_widgets()
        self._create_search_widgets()
        self._create_route_widgets()

    def _create_input_widgets(self) -> None:
        self._input["smiles"] = Text(description="SMILES", continuous_update=False)
        self._input["smiles"].observe(self._show_mol, names="value")
        display(self._input["smiles"])
        self._output["smiles"] = Output(
            layout={"border": "1px solid silver", "width": "50%", "height": "180px"}
        )
        display(self._output["smiles"])

        self._input["stocks"] = [
            Checkbox(
                value=True,
                description=key,
                style={"description_width": "initial"},
                layout={"justify": "left"},
            )
            for key in self.finder.stock.items
        ]

        list_ = [Label("Limit atom occurrences")]
        self._input["stocks_atom_count_on"] = []
        self._input["stocks_atom_count"] = []
        current_criteria = self.finder.stock.stop_criteria.get("counts", {})
        for atom in ["C", "O", "N"]:
            chk_box = Checkbox(
                value=atom in current_criteria,
                description=atom,
                layout={"justify": "left", "width": "80px"},
                style={"description_width": "initial"},
            )
            self._input["stocks_atom_count_on"].append(chk_box)
            inpt = BoundedIntText(
                value=current_criteria.get(atom, 0),
                min=0,
                layout={"width": "80px"},
            )
            self._input["stocks_atom_count"].append(inpt)
            list_.append(HBox([chk_box, inpt]))
        box_stocks = VBox(
            [Label("Stocks")] + self._input["stocks"] + list_,
            layout={"border": "1px solid silver"},
        )

        init_value = (
            [self.finder.expansion_policy.items[0]]
            if self.finder.expansion_policy
            else []
        )
        self._input["policy"] = SelectMultiple(
            options=self.finder.expansion_policy.items,
            value=init_value,
            description="Expansion Policy:",
            style={"description_width": "initial"},
            rows=min(len(self.finder.expansion_policy.items) + 1, 4),
        )

        self._input["filter"] = SelectMultiple(
            options=self.finder.filter_policy.items,
            description="Filter Policy:",
            style={"description_width": "initial"},
            rows=min(len(self.finder.filter_policy.items) + 1, 4),
        )

        max_time_box = self._make_slider_input("time_limit", "Time (min)", 1, 120)
        self._input["time_limit"].value = self.finder.config.search.time_limit / 60
        max_iter_box = self._make_slider_input(
            "iteration_limit", "Max Iterations", 100, 2000
        )
        self._input["iteration_limit"].value = self.finder.config.search.iteration_limit
        self._input["return_first"] = widgets.Checkbox(
            value=self.finder.config.search.return_first,
            description="Return first solved route",
        )
        self._input["max_transforms"] = BoundedIntText(
            description="Max tree depth",
            min=1,
            max=20,
            value=self.finder.config.search.max_transforms,
            style={"description_width": "initial"},
        )

        rewards = self.finder.config.search.algorithm_config.get("search_rewards", [])
        objective1 = rewards[0] if len(rewards) > 0 else "state score"
        objective2 = rewards[1] if len(rewards) > 1 else "None"
        self._input["reward1"] = widgets.Dropdown(
            options=self.finder.scorers.names(),
            description="MCTS reward 1:",
            value=objective1,
            style={"description_width": "initial"},
        )

        self._input["reward2"] = widgets.Dropdown(
            options=["None"] + self.finder.scorers.names(),
            description="MCTS reward 2:",
            value=objective2,
            style={"description_width": "initial"},
        )

        vbox = VBox(
            [
                self._input["policy"],
                self._input["filter"],
                max_time_box,
                max_iter_box,
                self._input["return_first"],
                self._input["max_transforms"],
                self._input["reward1"],
                self._input["reward2"],
            ]
        )
        box_options = HBox([box_stocks, vbox])
        display(box_options)

    def _create_route_widgets(self) -> None:
        self._input["scorer"] = widgets.Dropdown(
            options=self.finder.scorers.names(),
            description="Reorder by:",
            style={"description_width": "initial"},
        )
        self._input["scorer"].disabled = True
        self._input["scorer"].observe(self._on_change_scorer)
        self._buttons["show_routes"] = Button(description="Show Routes")
        self._buttons["show_routes"].on_click(self._on_display_button_clicked)
        self._input["route"] = Dropdown(
            options=[],
            description="Routes: ",
        )
        self._input["route"].disabled = True
        self._input["route"].observe(self._on_change_route_option)
        display(
            HBox(
                [
                    self._buttons["show_routes"],
                    self._input["route"],
                    self._input["scorer"],
                ]
            )
        )

        self._output["pareto_fronts"] = widgets.Output(
            layout={"border": "1px solid silver", "width": "99%", "overflow": "auto"}
        )
        display(self._output["pareto_fronts"])
        self._output["routes"] = widgets.Output(
            layout={"border": "1px solid silver", "width": "99%"}
        )
        display(self._output["routes"])

    def _create_search_widgets(self) -> None:
        self._buttons["execute"] = Button(description="Run Search")
        self._buttons["execute"].on_click(self._on_exec_button_clicked)

        self._buttons["extend"] = widgets.Button(description="Extend Search")
        self._buttons["extend"].on_click(self._on_extend_button_clicked)
        display(HBox([self._buttons["execute"], self._buttons["extend"]]))

        self._output["tree_search"] = widgets.Output(
            layout={
                "border": "1px solid silver",
                "width": "99%",
                "height": "320px",
                "overflow": "auto",
            }
        )
        display(self._output["tree_search"])

    def _make_slider_input(self, label, description, min_val, max_val) -> HBox:
        label_widget = Label(description)
        slider = widgets.IntSlider(
            continuous_update=True, min=min_val, max=max_val, readout=False
        )
        self._input[label] = IntText(continuous_update=True, layout={"width": "80px"})
        widgets.link((self._input[label], "value"), (slider, "value"))
        return HBox([label_widget, slider, self._input[label]])

    def _on_change_route_option(self, change) -> None:
        if change["name"] != "index":
            return
        self._show_route(self._input["route"].index)

    def _on_change_scorer(self, change) -> None:
        if self.finder.routes is None or change["name"] != "index":
            return
        scorer = self.finder.scorers[self._input["scorer"].value]
        self.finder.routes.rescore(scorer)
        self._show_route(self._input["route"].index)

    def _on_exec_button_clicked(self, _) -> None:
        self._toggle_button(False)
        self._prepare_search()
        self._tree_search()
        self._toggle_button(True)

    def _on_extend_button_clicked(self, _) -> None:
        self._toggle_button(False)
        self._tree_search()
        self._toggle_button(True)

    def _on_display_button_clicked(self, _) -> None:
        self._toggle_button(False)
        rewards = self.finder.config.search.algorithm_config["search_rewards"]
        self.finder.build_routes(scorer=rewards)
        self.finder.routes.make_images()
        self.finder.routes.compute_scores(*self.finder.scorers.objects())
        self._input["route"].options = [
            f"Option {i}" for i, _ in enumerate(self.finder.routes, 1)  # type: ignore
        ]

        self._output["pareto_fronts"].clear_output()
        if len(rewards) > 1:
            with self._output["pareto_fronts"]:
                pareto_fronts_plot(self.finder.routes)

        self._show_route(0)
        self._input["scorer"].disabled = len(rewards) > 1
        self._input["route"].disabled = False
        self._toggle_button(True)

    def _prepare_search(self) -> None:
        self._output["tree_search"].clear_output()
        with self._output["tree_search"]:
            selected_stocks = [
                cb.description for cb in self._input["stocks"] if cb.value
            ]
            self.finder.stock.select(selected_stocks)
            atom_count_limits = {}
            for cb_input, value_input in zip(
                self._input["stocks_atom_count_on"], self._input["stocks_atom_count"]
            ):
                if cb_input.value:
                    atom_count_limits[cb_input.description] = value_input.value
            self.finder.stock.set_stop_criteria({"counts": atom_count_limits})
            self.finder.expansion_policy.select(self._input["policy"].value)
            if not self._input["filter"].value:
                self.finder.filter_policy.deselect()
            else:
                self.finder.filter_policy.select(self._input["filter"].value)

            self.finder.config.search.max_transforms = self._input[
                "max_transforms"
            ].value
            self.finder.config.search.return_first = self._input["return_first"].value
            self.finder.config.search.time_limit = self._input["time_limit"].value * 60
            self.finder.config.search.iteration_limit = self._input[
                "iteration_limit"
            ].value

            rewards = [self._input["reward1"].value]
            if self._input["reward2"].value != "None":
                rewards.append(self._input["reward2"].value)
            self.finder.config.search.algorithm_config["search_rewards"] = rewards
            if self.finder.config.search.algorithm != "mcts":
                print(
                    "Only MCTS algorithm is supported in GUI interface. Resetting to MCTS"
                )
                self.finder.config.search.algorithm = "mcts"

            smiles = self._input["smiles"].value
            print(f"Setting target molecule with smiles: {smiles}")
            self.finder.target_smiles = smiles
            self.finder.prepare_tree()

    def _show_mol(self, change) -> None:
        self._output["smiles"].clear_output()
        with self._output["smiles"]:
            mol = Chem.MolFromSmiles(change["new"])
            display(mol)

    def _show_route(self, index) -> None:
        route_display(index, self.finder.routes, self._output["routes"])

    def _toggle_button(self, on_) -> None:
        for button in self._buttons.values():
            button.disabled = not on_

    def _tree_search(self) -> None:
        with self._output["tree_search"]:
            self.finder.tree_search(show_progress=True)
            display(HTML("<b>Tree search completed!</b>"))


def _get_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser("aizynthapp")
    parser.add_argument(
        "--config", required=True, help="the filename of a configuration file"
    )
    parser.add_argument("--output", help="the filename of the Jupyter notebook")
    return parser.parse_args()


def main() -> None:
    """Entry point for the aizynthapp command"""
    args = _get_arguments()

    commands = "\n".join(
        [
            "from aizynthfinder.interfaces import AiZynthApp",
            f'configfile=r"{os.path.abspath(args.config)}"',
            "app = AiZynthApp(configfile)",
        ]
    )
    notebook = jupytext.reads(commands, fmt="py:percent")
    if args.output:
        filename = args.output
    else:
        _, filename = tempfile.mkstemp(suffix=".ipynb")
    jupytext.write(notebook, filename, fmt="ipynb")

    if args.output:
        print(f"Notebook saved to {filename}. It can now be open with Jupyter notebook")
        return

    try:
        proc = subprocess.Popen(f"jupyter notebook {filename}".split())
        proc.communicate()
    except KeyboardInterrupt:
        proc.send_signal(signal.SIGINT)


if __name__ == "__main__":
    main()



""" Sub-package containing scoring routines
"""

from aizynthfinder.context.scoring.collection import ScorerCollection
from aizynthfinder.context.scoring.scorers import (
    AverageTemplateOccurrenceScorer,
    BrokenBondsScorer,
    CombinedScorer,
    DeltaSyntheticComplexityScorer,
    FractionInStockScorer,
    MaxTransformScorerer,
    NumberOfPrecursorsInStockScorer,
    NumberOfPrecursorsScorer,
    NumberOfReactionsScorer,
    PriceSumScorer,
    ReactionClassMembershipScorer,
    RouteCostScorer,
    RouteSimilarityScorer,
    Scorer,
    StateScorer,
    StockAvailabilityScorer,
    SUPPORT_DISTANCES,
)
from aizynthfinder.utils.exceptions import ScorerException



""" Module containing helper routines for using Keras, Tensorflow and Onnx models
"""
from __future__ import annotations

import functools
import logging
import os
from typing import TYPE_CHECKING

import numpy as np
import onnxruntime
import psutil
import requests

try:
    import grpc
    import tensorflow as tf
    from google.protobuf.json_format import MessageToDict

    # pylint: disable=all
    from tensorflow.keras.metrics import top_k_categorical_accuracy
    from tensorflow.keras.models import load_model as load_keras_model
    from tensorflow_serving.apis import (
        get_model_metadata_pb2,
        predict_pb2,
        prediction_service_pb2_grpc,
    )
except ImportError:
    SUPPORT_EXTERNAL_APIS = False
else:
    SUPPORT_EXTERNAL_APIS = True


from aizynthfinder.utils.exceptions import ExternalModelAPIError

# pylint: enable=all
from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import Any, Callable, List, Union

    _ModelInput = Union[np.ndarray, List[np.ndarray]]

_logger = logger()

# Suppress tensforflow logging
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

TF_SERVING_HOST = os.environ.get("TF_SERVING_HOST")
TF_SERVING_REST_PORT = os.environ.get("TF_SERVING_REST_PORT")
TF_SERVING_GRPC_PORT = os.environ.get("TF_SERVING_GRPC_PORT")


def load_model(
    source: str, key: str, use_remote_models: bool
) -> Union[
    "LocalKerasModel", "LocalOnnxModel", "ExternalModelViaGRPC", "ExternalModelViaREST"
]:
    """
    Load model from a configuration specification.

    If `use_remote_models` is True, tries to load:
      1. A Tensorflow server through gRPC
      2. A Tensorflow server through REST API
      3. A local Keras model
    otherwise it just loads the local model.

    :param source: if fallbacks to a local model, this is the filename
    :param key: when connecting to Tensorflow server this is the model name
    :param use_remote_models: if True will try to connect to remote model server
    :return: a model object with a predict object
    """
    if source.split(".")[-1] == "onnx":
        return LocalOnnxModel(source)

    if not SUPPORT_EXTERNAL_APIS:
        raise ValueError(
            "Tensorflow is not installed - this installation only supports ONNX models"
        )

    if not use_remote_models:
        return LocalKerasModel(source)

    try:
        return ExternalModelViaGRPC(key)
    except ExternalModelAPIError:
        pass
    try:
        return ExternalModelViaREST(key)
    except ExternalModelAPIError:
        pass
    return LocalKerasModel(source)


class LocalKerasModel:
    """
    A keras policy model that is executed locally.

    The size of the input vector can be determined with the len() method.

    :ivar model: the compiled model
    :ivar output_size: the length of the output vector

    :param filename: the path to a Keras checkpoint file
    """

    def __init__(self, filename: str) -> None:
        top10_acc = functools.partial(top_k_categorical_accuracy, k=10)
        top10_acc.__name__ = "top10_acc"  # type: ignore

        top50_acc = functools.partial(top_k_categorical_accuracy, k=50)
        top50_acc.__name__ = "top50_acc"  # type: ignore

        self.model = load_keras_model(
            filename,
            custom_objects={"top10_acc": top10_acc, "top50_acc": top50_acc, "tf": tf},
        )
        try:
            self._model_dimensions = int(self.model.input.shape[1])
        except AttributeError:
            self._model_dimensions = int(self.model.input[0].shape[1])
        self.output_size = int(self.model.output.shape[1])

    def __len__(self) -> int:
        return self._model_dimensions

    def predict(self, *args: np.ndarray, **_: np.ndarray) -> np.ndarray:
        """
        Perform a forward pass of the neural network.

        :param args: the input vectors
        :return: the vector of the output layer
        """
        return self.model.predict(args, verbose=0)


class LocalOnnxModel:
    """
    An Onnx model that is executed locally.

    The size of the input vector can be determined with the len() method.

    :ivar model: the compiled Onnx model
    :ivar output_size: the length of the output vector

    :param filename: the path to a Onnx model checkpoint file
    """

    def __init__(self, filename: str) -> None:
        session_options = onnxruntime.SessionOptions()
        session_options.intra_op_num_threads = _get_thread_count_per_core()

        self.model = onnxruntime.InferenceSession(
            filename, sess_options=session_options
        )
        self._model_inputs = self.model.get_inputs()
        self._model_output = self.model.get_outputs()[0]
        self._model_dimensions = int(self._model_inputs[0].shape[1])
        self.output_size = int(self._model_output.shape[1])

    def __len__(self) -> int:
        return self._model_dimensions

    def predict(self, *args: np.ndarray, **_: np.ndarray) -> np.ndarray:
        """
        Perform a prediction run on the onnx model.

        :param args: the input vectors
        :return: the vector of the output layer
        """
        return self.model.run(
            [self._model_output.name],
            {
                model_input.name: input.astype(np.float32)
                for model_input, input in zip(self._model_inputs, list(args))
            },
        )[0]


def _log_and_reraise_exceptions(method: Callable) -> Callable:
    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except Exception as err:
            msg = "Error when requesting from tensorflow model API"
            _logger.error("%s: %s", msg, err)
            raise ExternalModelAPIError(msg)

    return wrapper


class ExternalModelViaREST:
    """
    A neural network model implementation using TF Serving via REST API.

    :param name: the name of model
    """

    def __init__(self, name: str) -> None:
        if not SUPPORT_EXTERNAL_APIS:
            raise ExternalModelAPIError("API packages are not installed.")

        self._model_url = self._get_model_url(name)
        self._sig_def = self._get_sig_def()

    def __len__(self) -> int:
        first_input_name = list(self._sig_def["inputs"].keys())[0]
        return int(
            self._sig_def["inputs"][first_input_name]["tensor_shape"]["dim"][1]["size"]
        )

    def predict(self, *args: np.ndarray, **kwargs: np.ndarray) -> np.ndarray:
        """
        Get prediction from model.

        If the keys in `kwargs` agree with the model input names, they
        will be used in stead of `arg`

        :param args: the input vectors
        :param kwargs: the named input vectors
        :return: the vector of the output layer
        """
        url = self._model_url + ":predict"
        res = self._handle_rest_api_request(
            "POST", url, json=self._make_payload(*args, **kwargs)
        )
        return np.asarray(res["outputs"])

    def _get_sig_def(self) -> dict:
        res = self._handle_rest_api_request("GET", self._model_url + "/metadata")
        return res["metadata"]["signature_def"]["signature_def"]["serving_default"]

    # pylint: disable=no-self-use
    @_log_and_reraise_exceptions
    def _handle_rest_api_request(
        self, method: str, url: str, *args: Any, **kwargs: Any
    ) -> dict:
        res = requests.request(method, url, *args, **kwargs)
        if res.status_code != 200 or (
            res.headers["Content-Type"] != "application/json"
        ):
            raise ExternalModelAPIError(
                f"Unexpected response from REST API: {res.status_code}\n{res.text}"
            )

        return res.json()

    def _make_payload(self, *args: np.ndarray, **kwargs: np.ndarray) -> dict:
        if all(key in kwargs for key in self._sig_def["inputs"].keys()):
            data = {key: kwargs[key].tolist() for key in self._sig_def["inputs"].keys()}
        else:
            data = {
                name: fp.tolist()
                for name, fp in zip(self._sig_def["inputs"].keys(), args)
            }
        return {"inputs": data}

    @staticmethod
    def _get_model_url(name: str) -> str:
        warning = f"Failed to get url of REST service for external model {name}"
        if not TF_SERVING_HOST:
            _logger.warning(warning)
            raise ExternalModelAPIError("Host not set for model {name}")
        if not TF_SERVING_REST_PORT:
            _logger.warning(warning)
            raise ExternalModelAPIError("REST port not set for model {name}")
        return f"http://{TF_SERVING_HOST}:{TF_SERVING_REST_PORT}/v1/models/{name}"


class ExternalModelViaGRPC:
    """
    A neural network model implementation using TF Serving via gRPC.

    :param name: the name of model
    """

    def __init__(self, name: str) -> None:
        if not SUPPORT_EXTERNAL_APIS:
            raise ExternalModelAPIError("API packages are not installed.")

        self._server = self._get_server(name)
        self._model_name = name
        self._sig_def = self._get_sig_def()

    def __len__(self) -> int:
        first_input_name = list(self._sig_def["inputs"].keys())[0]
        return int(
            self._sig_def["inputs"][first_input_name]["tensorShape"]["dim"][1]["size"]
        )

    @_log_and_reraise_exceptions
    def predict(self, *args: np.ndarray, **kwargs: np.ndarray) -> np.ndarray:
        """
        Get prediction from model.

        If the keys in `kwargs` agree with the model input names, they
        will be used in stead of `arg`

        :param args: the input vectors
        :param kwargs: the named input vectors
        :return: the vector of the output layer
        """
        input_tensors = self._make_payload(*args, **kwargs)
        channel = grpc.insecure_channel(self._server)
        service = prediction_service_pb2_grpc.PredictionServiceStub(channel)
        request = predict_pb2.PredictRequest()
        request.model_spec.name = self._model_name
        for name, tensor in input_tensors.items():
            request.inputs[name].CopyFrom(tensor)
        key = list(self._sig_def["outputs"].keys())[0]
        return tf.make_ndarray(service.Predict(request, 10.0).outputs[key])

    @_log_and_reraise_exceptions
    def _get_sig_def(self) -> dict:
        channel = grpc.insecure_channel(self._server)
        service = prediction_service_pb2_grpc.PredictionServiceStub(channel)
        request = get_model_metadata_pb2.GetModelMetadataRequest()
        request.model_spec.name = self._model_name
        request.metadata_field.append("signature_def")
        result = MessageToDict(service.GetModelMetadata(request, 10.0))
        # close the channel so that it won't be reused after fork and fail
        channel.close()
        return result["metadata"]["signature_def"]["signatureDef"]["serving_default"]

    def _make_payload(self, *args: np.ndarray, **kwargs: np.ndarray) -> dict:
        if all(key in kwargs for key in self._sig_def["inputs"].keys()):
            inputs = kwargs
        else:
            inputs = dict(zip(self._sig_def["inputs"].keys(), args))

        tensors = {}
        for name, vec in inputs.items():
            size = int(self._sig_def["inputs"][name]["tensorShape"]["dim"][1]["size"])
            assert size == vec.shape[1], "Incorrect shape for input"
            tensors[name] = tf.make_tensor_proto(vec, dtype=np.float32, shape=vec.shape)
        return tensors

    @staticmethod
    def _get_server(name: str) -> str:
        warning = f"Failed to get gRPC server for external model {name}"
        if not TF_SERVING_HOST:
            _logger.warning(warning)
            raise ExternalModelAPIError(f"Host not set for model {name}")
        if not TF_SERVING_GRPC_PORT:
            _logger.warning(warning)
            raise ExternalModelAPIError(f"GRPC port not set for model {name}")
        return f"{TF_SERVING_HOST}:{TF_SERVING_GRPC_PORT}"


def _get_thread_count_per_core() -> int:
    return psutil.cpu_count() // psutil.cpu_count(logical=False)



""" Module containing classes used to score the reaction routes.
"""

from __future__ import annotations

import abc
import json
from collections import defaultdict
from collections.abc import Sequence as SequenceAbc
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

try:
    from route_distances.route_distances import route_distances_calculator
except ImportError:
    SUPPORT_DISTANCES = False
else:
    SUPPORT_DISTANCES = True

from aizynthfinder.chem import TreeMolecule
from aizynthfinder.context.stock import StockException
from aizynthfinder.reactiontree import ReactionTree
from aizynthfinder.search.mcts import MctsNode
from aizynthfinder.utils.bonds import BrokenBonds
from aizynthfinder.utils.exceptions import ScorerException
from aizynthfinder.utils.logging import logger
from aizynthfinder.utils.sc_score import SCScore

if TYPE_CHECKING:
    from aizynthfinder.chem import (
        FixedRetroReaction,
        Molecule,
        RetroReaction,
        UniqueMolecule,
    )
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import (
        Iterable,
        Optional,
        Sequence,
        StrDict,
        Tuple,
        TypeVar,
        Union,
    )

    _Molecules = Sequence[Molecule]
    _Scoreable = TypeVar("_Scoreable", MctsNode, ReactionTree)
    _Scoreables = Sequence[_Scoreable]
    _ScorerItemType = Union[_Scoreables, _Scoreable]


@dataclass
class SquashScaler:
    """
    Squash function loosely adapted from a sigmoid function with parameters
    to modify and offset the shape

    :param slope: the slope of the midpoint
    :param xoffset: the offset of the midpoint along the x-axis
    :param yoffset: the offset of the curve along the y-axis
    """

    slope: float
    xoffset: float
    yoffset: float

    def __call__(self, val: float) -> float:
        return 1 / (1 + np.exp(self.slope * -(val - self.xoffset))) - self.yoffset


@dataclass
class MinMaxScaler:
    """
    Scaling function that normalises the value between 0 - 1,
    the reverse variable controls the direction of scaling,
    reverse should set to be true for rewards that need to be minimised
    the scale_factor could be used to adjust the scores when they are too small or too big

    :param val: the value that is being scaled
    :param min_val: minimum val param val could take
    :param max_val: maximum val param val could take
    :param scale_factor: scaling factor applied to the minmax scaled output
    """

    min_val: float
    max_val: float
    reverse: bool
    scale_factor: float = 1

    def __call__(self, val: float) -> float:
        val = np.clip(val, self.min_val, self.max_val)
        if self.reverse:
            numerator = self.max_val - val
        else:
            numerator = val - self.min_val
        return (numerator / (self.max_val - self.min_val)) * self.scale_factor


_SCALERS = {"squash": SquashScaler, "min_max": MinMaxScaler}


class Scorer(abc.ABC):
    """
    Abstract base class for classes that do scoring on MCTS-like nodes or reaction trees.

    The actual scoring is done be calling an instance of
    a scorer class with a ``Node`` or ``ReactionTree`` object as only argument.

    .. code-block::

        scorer = MyScorer()
        score = scorer(node1)

    You can also give a list of such objects to the scorer

    .. code-block::

        scorer = MyScorer()
        scores = scorer([node1, node2])

    :param config: the configuration the tree search
    :param scaler_params: the parameter settings of the scaler
    """

    scorer_name = "base"

    def __init__(
        self,
        config: Optional[Configuration] = None,
        scaler_params: Optional[StrDict] = None,
    ) -> None:
        self._config = config
        self._reverse_order: bool = True
        self._scaler = None
        self._scaler_name = ""
        if scaler_params:
            self._scaler_name = scaler_params["name"]
            del scaler_params["name"]
            if scaler_params:
                self._scaler = _SCALERS[self._scaler_name](**scaler_params)
            else:
                # for paramterless function
                self._scaler = _SCALERS[self._scaler_name]

    def __call__(self, item: _ScorerItemType) -> Union[float, Sequence[float]]:
        if isinstance(item, SequenceAbc):
            return self._score_many(item)
        if isinstance(item, (MctsNode, ReactionTree)):
            return self._score_just_one(item)  # type: ignore
        raise ScorerException(
            f"Unable to score item from class {item.__class__.__name__}"
        )

    def __repr__(self) -> str:
        repr_name = self.scorer_name
        if self._scaler_name:
            repr_name += f"-{self._scaler_name}"
        return repr_name

    def sort(
        self, items: _Scoreables
    ) -> Tuple[_Scoreables, Sequence[float], Sequence[int]]:
        """
        Sort nodes or reaction trees in descending order based on the score

        :param items: the items to sort
        :return: the sorted items and their scores
        """
        scores = self._score_many(items)
        assert isinstance(scores, SequenceAbc)
        sortidx = sorted(
            range(len(scores)), key=scores.__getitem__, reverse=self._reverse_order
        )
        scores = [scores[idx] for idx in sortidx]
        sorted_items = [items[idx] for idx in sortidx]
        return sorted_items, scores, sortidx

    def _score_just_one(self, item: _Scoreable) -> float:
        if isinstance(item, MctsNode):
            node_score = self._score_node(item)
            if self._scaler:
                node_score = self._scaler(node_score)
            return node_score
        if isinstance(item, ReactionTree):
            tree_score = self._score_reaction_tree(item)
            if self._scaler:
                tree_score = self._scaler(tree_score)
            return tree_score
        raise ScorerException(
            f"Unable to score item from class {item.__class__.__name__}"
        )

    def _score_many(self, items: _Scoreables) -> Sequence[float]:
        return [self._score_just_one(item) for item in items]

    @abc.abstractmethod
    def _score_node(self, node: MctsNode) -> float:
        pass

    @abc.abstractmethod
    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        pass


class StateScorer(Scorer):
    """Class for scoring nodes based on the state score"""

    scorer_name = "state score"

    def __init__(
        self, config: Configuration, scaler_params: Optional[StrDict] = None
    ) -> None:
        super().__init__(config, scaler_params)
        # This is necessary because config should not be optional for this scorer
        self._config: Configuration = config
        self._transform_scorer = MaxTransformScorerer(
            config,
            scaler_params={"name": "squash", "slope": -1, "yoffset": 0, "xoffset": 4},
        )
        self._in_stock_scorer = FractionInStockScorer(config)

    def _score(self, item: _Scoreable) -> float:
        in_stock_fraction = self._in_stock_scorer(item)
        max_transform = self._transform_scorer(item)
        # A scorer can return a list of float if the item is a list of trees/nodes,
        # but that is not the case here. However this is needed because of mypy
        assert isinstance(in_stock_fraction, float) and isinstance(max_transform, float)
        return 0.95 * in_stock_fraction + 0.05 * max_transform

    def _score_node(self, node: MctsNode) -> float:
        return self._score(node)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        return self._score(tree)


class MaxTransformScorerer(Scorer):
    """Class for scoring nodes based on the maximum transform"""

    scorer_name = "max transform"

    def _score_node(self, node: MctsNode) -> float:
        return node.state.max_transforms

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        mols = [
            TreeMolecule(
                parent=None, transform=tree.depth(leaf) // 2, smiles=leaf.smiles
            )
            for leaf in tree.leafs()
        ]
        return max(mol.transform for mol in mols)


class FractionInStockScorer(Scorer):
    """Class for scoring nodes based on the fraction in stock"""

    scorer_name = "fraction in stock"

    def __init__(
        self, config: Configuration, scaler_params: Optional[StrDict] = None
    ) -> None:
        super().__init__(config, scaler_params)
        # This is necessary because config should not be optional for this scorer
        self._config: Configuration = config

    def _score_node(self, node: MctsNode) -> float:
        num_in_stock = np.sum(node.state.in_stock_list)
        num_molecules = len(node.state.mols)
        return float(num_in_stock) / float(num_molecules)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        leaves = list(tree.leafs())
        num_in_stock = sum(mol in self._config.stock for mol in leaves)
        num_molecules = len(leaves)
        return float(num_in_stock) / float(num_molecules)


class NumberOfReactionsScorer(Scorer):
    """Class for scoring nodes based on the number of reaction it took to get to a node"""

    scorer_name = "number of reactions"

    def __init__(
        self,
        config: Optional[Configuration] = None,
        scaler_params: Optional[StrDict] = None,
    ) -> None:
        super().__init__(config, scaler_params)
        self._reverse_order = False

    def _score_node(self, node: MctsNode) -> float:
        reactions = node.actions_to()
        return len(reactions)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        return len(list(tree.reactions()))


class NumberOfPrecursorsScorer(Scorer):
    """Class for scoring nodes based on the number of pre-cursors in a node or route"""

    scorer_name = "number of pre-cursors"

    def __init__(
        self,
        config: Optional[Configuration] = None,
        scaler_params: Optional[StrDict] = None,
    ) -> None:
        super().__init__(config, scaler_params)
        self._reverse_order = False

    def _score_node(self, node: MctsNode) -> float:
        return len(node.state.mols)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        return len(list(tree.leafs()))


class NumberOfPrecursorsInStockScorer(Scorer):
    """Class for scoring nodes based on the number of pre-cursors in stock a
    node or route"""

    scorer_name = "number of pre-cursors in stock"

    def __init__(
        self, config: Configuration, scaler_params: Optional[StrDict] = None
    ) -> None:
        super().__init__(config, scaler_params)
        self._stock = config.stock

    def _score_node(self, node: MctsNode) -> float:
        return len([mol for mol in node.state.mols if mol in self._stock])

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        return len([mol for mol in tree.leafs() if mol in self._stock])


class AverageTemplateOccurrenceScorer(Scorer):
    """Class for scoring the nodes based on the average occurrence of the
    templates used to get to a node"""

    scorer_name = "average template occurrence"

    def _calc_average(
        self, reactions: Sequence[Union[FixedRetroReaction, RetroReaction]]
    ) -> float:
        if not reactions:
            return 0.0
        occurrences = [self._get_occurrence(reaction) for reaction in reactions]
        return sum(occurrences) / len(reactions)

    def _score_node(self, node: MctsNode) -> float:
        return self._calc_average(node.actions_to())

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        return self._calc_average(list(tree.reactions()))

    @staticmethod
    def _get_occurrence(reaction: Union[FixedRetroReaction, RetroReaction]) -> int:
        return reaction.metadata.get(
            "library_occurrence", reaction.metadata.get("library_occurence", 0)
        )


class PriceSumScorer(Scorer):
    """Scorer that sums the prices of all pre-cursors"""

    scorer_name = "sum of prices"

    def __init__(
        self,
        config: Configuration,
        scaler_params: Optional[StrDict] = None,
        default_cost: float = 1.0,
        not_in_stock_multiplier: int = 10,
    ) -> None:
        super().__init__(config, scaler_params)
        self._config: Configuration = config
        self.default_cost = default_cost
        self.not_in_stock_multiplier = not_in_stock_multiplier
        self._reverse_order = False

    def _calculate_leaf_costs(
        self, leafs: Union[Sequence[Molecule], Iterable[Molecule]]
    ) -> dict:
        costs = {}
        for mol in leafs:
            if mol not in self._config.stock:
                continue
            try:
                cost = self._config.stock.price(mol)
            except StockException:
                costs[mol] = self.default_cost
            else:
                costs[mol] = cost

        max_cost = max(costs.values()) if costs else self.default_cost
        return defaultdict(lambda: max_cost * self.not_in_stock_multiplier, costs)

    def _score_node(self, node: MctsNode) -> float:
        leaf_costs = self._calculate_leaf_costs(node.state.mols)
        return sum(leaf_costs[mol] for mol in node.state.mols)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        leaf_costs = self._calculate_leaf_costs(tree.leafs())
        return sum(leaf_costs[leaf] for leaf in tree.leafs())


class RouteCostScorer(PriceSumScorer):
    """
    Score based on the cost of molecules and reactions.
    From Badowski et al. Chem Sci. 2019, 10, 4640
    """

    scorer_name = "route cost"

    def __init__(
        self,
        config: Configuration,
        scaler_params: Optional[StrDict] = None,
        reaction_cost: int = 1,
        average_yield: float = 0.8,
        default_cost: int = 1,
        not_in_stock_multiplier: int = 10,
    ) -> None:
        super().__init__(
            config,
            scaler_params,
            default_cost=default_cost,
            not_in_stock_multiplier=not_in_stock_multiplier,
        )
        self.reaction_cost = reaction_cost
        self.average_yield = average_yield
        self._reverse_order = False

    def _score_node(self, node: MctsNode) -> float:
        leaf_costs = self._calculate_leaf_costs(node.state.mols)

        reactions, nodes = node.path_to()
        if not reactions:
            return leaf_costs[node.state.mols[0]]

        scores = {id(mol): leaf_costs[mol] for mol in nodes[-1].state.mols}
        for pnode, reaction in zip(nodes[::-1][1:], reactions[::-1]):
            updated_scores = {
                id(mol): scores[id(mol)]
                for mol in pnode.state.mols
                if mol is not reaction.mol
            }
            child_sum = sum(
                1 / self.average_yield * score
                for id_, score in scores.items()
                if id_ not in updated_scores
            )
            updated_scores[id(reaction.mol)] = self.reaction_cost + child_sum
            scores = updated_scores

        return list(scores.values())[0]

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        def _recursive_score(node):
            # This list should contains 0 or 1 elements
            reaction_nodes = list(tree.graph[node])
            if not reaction_nodes:
                return leaf_costs[node]

            child_sum = sum(
                1 / self.average_yield * _recursive_score(child)
                for child in tree.graph[reaction_nodes[0]]
            )
            return self.reaction_cost + child_sum

        leaf_costs = self._calculate_leaf_costs(tree.leafs())
        return _recursive_score(tree.root)


class ReactionClassMembershipScorer(Scorer):
    """
    Scorer that checks if the reaction classes are in a specified set

    The score is calculated as product over each reaction. For each reaction
    the reaction classification is checked if it is in a given list of classes.
    """

    def __init__(
        self,
        config: Configuration,
        reaction_class_set: Sequence[str],
        in_set_score: float = 1.0,
        not_in_set_score: float = 0.1,
    ) -> None:
        super().__init__(config)
        self.reaction_class_set = set(item.split(" ")[0] for item in reaction_class_set)
        self.in_set_score = in_set_score
        self.not_in_set_score = not_in_set_score

    def __repr__(self) -> str:
        return "reaction class membership"

    def _calc_product(
        self, reactions: Sequence[Union[FixedRetroReaction, RetroReaction]]
    ) -> float:
        if not reactions:
            return 1.0
        prod = 1.0
        for reaction in reactions:
            membership = self._get_membership(reaction)
            prod *= self.in_set_score if membership else self.not_in_set_score
        return prod

    def _get_membership(
        self, reaction: Union[FixedRetroReaction, RetroReaction]
    ) -> bool:
        classification = reaction.metadata.get("classification")
        if not classification:
            return 0.0 in self.reaction_class_set
        return classification.split(" ")[0] in self.reaction_class_set

    def _score_node(self, node: MctsNode) -> float:
        return self._calc_product(node.actions_to())

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        return self._calc_product(list(tree.reactions()))


class StockAvailabilityScorer(Scorer):
    """
    Scorer that computes score based on the stock availability of the starting material

    The score is calculated as a product of a stock score per starting material. The stock
    score for each molecule is based on the source of the stock, or a default value if the
    molecule is not in stock. The `other_source_score` parameter can be used to
    distinguish between "not in stock" and "not in the specificed sources" cases.
    """

    def __init__(
        self,
        config: Configuration,
        source_score: StrDict,
        default_score: float = 0.1,
        other_source_score: Optional[float] = None,
    ) -> None:
        super().__init__(config)
        assert self._config is not None
        self.source_score = source_score
        self.default_score = default_score
        self.other_source_score = other_source_score

    def __repr__(self) -> str:
        return "stock availability"

    def _calculate_leaf_costs(
        self, leafs: Union[Sequence[Molecule], Iterable[Molecule]]
    ) -> float:
        assert self._config is not None
        prod = 1.0
        for mol in leafs:
            availability = self._config.stock.availability_list(mol)
            scores = [
                self.source_score[source]
                for source in availability
                if source in self.source_score
            ]
            if scores:
                prod *= max(scores)
            elif self.other_source_score and mol in self._config.stock:
                prod *= self.other_source_score
            else:
                prod *= self.default_score
        return prod

    def _score_node(self, node: MctsNode) -> float:
        return self._calculate_leaf_costs(node.state.mols)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        return self._calculate_leaf_costs(tree.leafs())


class BrokenBondsScorer(Scorer):
    """Class for scoring nodes and reaction trees based on the breaking of atom bonds

    The score is a summation of the depths in the tree where the focussed bonds
    are found to break in the reaction. If a focussed bond is found to be unbroken
    in the entire tree, the total length of the tree will be added to the score.
    """

    scorer_name = "broken bonds"

    def __init__(self, config: Configuration) -> None:
        super().__init__(config)
        self._break_bonds = [tuple(bond) for bond in config.search.break_bonds]
        self._break_bonds_operator = config.search.break_bonds_operator.lower()
        self._reverse_order = False

    def __repr__(self) -> str:
        return self.scorer_name

    def _calculate_broken_bonds_score(
        self, reactions: Sequence[RetroReaction], depths: Sequence[int]
    ) -> float:

        if not reactions:
            return 0

        max_score = len(set(depths)) * len(self._break_bonds)
        broken_focussed_bonds = []
        scores = []

        # The score should be 0 when no transformations were reported
        if not depths:
            return 0

        for reaction, depth in zip(reactions, depths):
            broken_bonds = BrokenBonds(self._break_bonds)(reaction)
            broken_untracked_bonds = [
                bond for bond in broken_bonds if bond not in broken_focussed_bonds
            ]
            if len(broken_untracked_bonds) > 0:
                broken_focussed_bonds += broken_untracked_bonds
                scores.append(1 - (depth / max_score))
                if self._break_bonds_operator == "or":
                    break

        if self._break_bonds_operator != "or" or (
            self._break_bonds_operator == "or" and len(broken_focussed_bonds) == 0
        ):
            # type: ignore
            if unbroken_bonds := set(self._break_bonds) - set(broken_focussed_bonds):
                scores.append(1 - (len(set(depths)) * len(unbroken_bonds)) / max_score)
        return sum(scores) / len(scores)

    def _score_node(self, node: MctsNode) -> float:
        reactions = node.actions_to()
        depths = []
        for reaction in reactions:
            depth = (
                max(
                    reactant.transform
                    for reactant in reaction.reactants[reaction.index]
                )
                - 1
            )
            depths.append(depth)
        return self._calculate_broken_bonds_score(reactions, depths)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        reactions = [
            reaction.to_smiles_based_retroreaction() for reaction in tree.reactions()
        ]
        depths = [tree.depth(reaction) // 2 for reaction in tree.reactions()]
        return self._calculate_broken_bonds_score(reactions, depths)


class RouteSimilarityScorer(Scorer):
    """
    Class for scoring based on an LSTM model for computing Tree Edit Distance to
    a set of reference routes.

    :param config: the configuration of the tree search
    :param routes_path: the filename of a JSON file with reference routes
    :param model_path: the filename of a checkpoint file with the LSTM model
    :param scaler_params: the parameter settings of the scaler
    :param agg_func: the name of numpy function used to aggregate the distances
                     to the reference routes
    :param similarity: if True, will compute similarity score else distance scores
    """

    scorer_name = "route similarity"

    def __init__(
        self,
        config: Configuration,
        routes_path: str,
        model_path: str,
        scaler_params: Optional[StrDict] = None,
        agg_func: str = "min",
        similarity: bool = False,
    ) -> None:
        if not SUPPORT_DISTANCES:
            raise ValueError(
                "Distance calculations are not supported by this installation."
                " Please install aizynthfinder with extras dependencies."
            )

        # Default scaler from benchmarking
        if scaler_params is None:
            scaler_params = {
                "name": "squash",
                "slope": 0.5,
                "xoffset": 10,
                "yoffset": 0,
            }

        super().__init__(config, scaler_params)
        self._reverse_order = similarity
        self.calculator = route_distances_calculator("lstm", model_path=model_path)

        # Scalar needs to be applied before taking into account of `similarity` parameter,
        # hence we are reassigning the global scaler and then disabling it.
        self._local_scaler = self._scaler
        self._scaler = None

        if not hasattr(np, agg_func):
            raise ValueError(f"Cannot identify aggregate function {agg_func} in numpy")
        self.agg_func = getattr(np, agg_func)
        self.similarity = similarity

        try:
            with open(routes_path or "") as file:
                self.routes = json.load(file)
        except FileNotFoundError:
            logger().info(
                f"Could not load reference routes from {routes_path}. Assuming they will be set later"
            )
            self.routes = []
        self.n_routes = len(self.routes)

    def _score_node(self, node: MctsNode) -> float:
        # We don't have any short-cut to score a node,
        # so we need to convert it to a reaction tree
        return self._score_reaction_tree(node.to_reaction_tree())

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        if not self.routes:
            return 0.0 if self.similarity else 1.0

        dist_matrix = self.calculator(self.routes + [tree.to_dict()])
        distances = dist_matrix[self.n_routes, : self.n_routes]
        score = self.agg_func(distances)
        norm_score = self._local_scaler(score)
        if self.similarity:
            return 1.0 - float(norm_score)
        return float(norm_score)


class DeltaSyntheticComplexityScorer(Scorer):
    """
    Class for scoring nodes based on the delta-synthetic-complexity of the node
    and its parent 'horizon' steps up in the tree.

    :param config: the configuration the tree search
    :param sc_score_model: the path to the SCScore model
    :param scaler_params: the parameter settings of the scaler, defaults to max-min between -1.5 and 4
    :param horizon: the number of steps backwards to look for parent molecule
    """

    scorer_name: str = "delta-SC score"

    def __init__(
        self,
        config: Configuration,
        sc_score_model: str,
        scaler_params: Optional[StrDict] = None,
        horizon: int = 3,
    ) -> None:
        if scaler_params is None:
            scaler_params = {
                "name": "min_max",
                "min_val": -1.5,
                "max_val": 4,
                "reverse": False,
            }
        super().__init__(config, scaler_params)
        ## This is necessary because config should not be optional for this scorer
        self._config: Configuration = config

        self.horizon = horizon
        self._model = SCScore(sc_score_model)

    def sc_deltas(self, mols: _Molecules, parents: _Molecules) -> Sequence[float]:
        """
        Calculate delta SC-score among the list of mol-parent pairs.

        :params mols: the leaves of the tree
        :param parents: the parent of the leaves
        :returns: the pair-wise difference in SCScore
        """
        delta_sc_scores = []
        for mol, parent in zip(mols, parents):
            mol.sanitize()
            parent.sanitize()

            sc_score_mol = self._model(mol.rd_mol)
            sc_score_parent = self._model(parent.rd_mol)

            delta_sc_score = sc_score_parent - sc_score_mol
            delta_sc_scores.append(delta_sc_score)
        return delta_sc_scores

    def _get_parent_from_reaction_tree(
        self, tree: ReactionTree, mol: UniqueMolecule, horizon: int
    ) -> UniqueMolecule:
        horizon = min(horizon, tree.depth(mol) // 2)

        parent = mol
        # Step up in the tree, first a reaction, then a molecule
        for _ in range(horizon):
            parent = tree.parent_molecule(parent)

        return parent

    def _get_parent_from_tree_molecule(
        self, mol: TreeMolecule, horizon: int
    ) -> TreeMolecule:
        horizon = min(horizon, mol.transform)

        parent = mol
        # Step up in the tree via parent nodes
        for _ in range(horizon):
            grandparent = parent.parent
            if not grandparent:
                break
            parent = grandparent

        return parent

    def _score_node(self, node: MctsNode) -> float:
        expandable_mols = node.state.expandable_mols
        if not expandable_mols:
            expandable_mols = list(node.state.mols)

        parent_mols = [
            self._get_parent_from_tree_molecule(mol, self.horizon)
            for mol in expandable_mols
        ]
        delta_sc_scores = self.sc_deltas(expandable_mols, parent_mols)
        return min(delta_sc_scores)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        expandable_mols = [node for node in tree.leafs() if not tree.in_stock(node)]
        if not expandable_mols:
            expandable_mols = list(tree.leafs())

        parent_mols = [
            self._get_parent_from_reaction_tree(tree, node, self.horizon)
            for node in expandable_mols
        ]
        delta_sc_scores = self.sc_deltas(expandable_mols, parent_mols)
        return min(delta_sc_scores)


class CombinedScorer(Scorer):
    """Class for scoring nodes and reaction trees by combining weighted scores from a list of scorers

    If no weights are provided as input, the scorer provides default weights that are
    equal for all input scorers.

    The CombinedScorer cannot be instantiated from the config file as it requires the
    names of the scorers to combine as input.
    """

    def __init__(
        self,
        config: Configuration,
        scorers: Sequence[str],
        weights: Optional[Sequence[float]] = None,
    ) -> None:
        super().__init__(config)
        self._scorers = [config.scorers[scorer] for scorer in scorers]
        self._weights = weights if weights else [1 / len(scorers)] * len(scorers)

    def __repr__(self) -> str:
        return " + ".join([repr(scorer) for scorer in self._scorers])

    def _combine_score(self, scores: Sequence[float]) -> float:
        return sum(
            score * weight for score, weight in zip(scores, self._weights)
        ) / sum(self._weights)

    def _score_node(self, node: MctsNode) -> float:
        scores = [scorer(node) for scorer in self._scorers]
        return self._combine_score(scores)

    def _score_reaction_tree(self, tree: ReactionTree) -> float:
        scores = [scorer(tree) for scorer in self._scorers]
        return self._combine_score(scores)



""" Sub-package containing analysis routines
"""
from aizynthfinder.analysis.tree_analysis import TreeAnalysis  # isort: skip
from aizynthfinder.analysis.routes import RouteCollection
from aizynthfinder.analysis.utils import RouteSelectionArguments



""" Module containing classes and routines for making stock input to the tree search.
"""
from __future__ import annotations

import argparse
import importlib
from typing import TYPE_CHECKING

try:
    import molbloom
except ImportError:
    HAS_MOLBLOOM = False
else:
    HAS_MOLBLOOM = True

import pandas as pd
from rdkit import Chem

from aizynthfinder.chem import Molecule, MoleculeException
from aizynthfinder.context.stock import MongoDbInchiKeyQuery

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import Iterable, List, Optional

    _StrIterator = Iterable[str]


def _get_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser("smiles2stock")
    parser.add_argument(
        "--files",
        required=True,
        nargs="+",
        help="the files containing smiles",
    )
    parser.add_argument(
        "--source",
        choices=["plain", "module"],
        help="indicates how to read the files. "
        "If 'plain' is used the input files should only contain SMILES (one on each row), "
        "if 'module' is used the SMILES are loaded from by python module"
        " (see documentation for details)",
        default="plain",
    )
    parser.add_argument(
        "--output",
        required=True,
        default="",
        help="the name of the output file or source tag",
    )
    parser.add_argument(
        "--target",
        choices=["hdf5", "mongo", "molbloom", "molbloom-inchi"],
        help="type of output",
        default="hdf5",
    )
    parser.add_argument("--host", help="the host of the Mongo database")
    parser.add_argument(
        "--bloom_params", nargs=2, type=int, help="the parameters to the Bloom filter"
    )
    return parser.parse_args()


def _convert_smiles(smiles_list: _StrIterator) -> _StrIterator:
    for smiles in smiles_list:
        try:
            yield Molecule(smiles=smiles, sanitize=True).inchi_key
        except MoleculeException:
            print(
                f"Failed to convert {smiles} to inchi key. Probably due to sanitation.",
                flush=True,
            )


def extract_plain_smiles(files: List[str]) -> _StrIterator:
    """
    Extract SMILES from plain text files, one SMILES on each line.
    The SMILES are yielded to save memory.
    """
    for filename in files:
        print(f"Processing {filename}", flush=True)
        with open(filename, "r") as fileobj:
            for line in fileobj:
                yield line.strip()


def extract_smiles_from_module(files: List[str]) -> _StrIterator:
    """
    Extract SMILES by loading a custom module, containing
    the function ``extract_smiles``.

    The first element of the input argument is taken as the module name.
    The other elements are taken as input to the ``extract_smiles`` method

    The SMILES are yielded to save memory.
    """
    module_name = files.pop(0)
    module = importlib.import_module(module_name)
    if not files:
        for smiles in module.extract_smiles():  # type: ignore  # pylint: disable=R1737
            yield smiles
    else:
        for filename in files:
            print(f"Processing {filename}", flush=True)
            for smiles in module.extract_smiles(filename):  # type: ignore  # pylint: disable=R1737
                yield smiles


def make_hdf5_stock(inchi_keys: _StrIterator, filename: str) -> None:
    """
    Put all the inchi keys from the given iterable in a pandas
    dataframe and save it as an HDF5 file. Only unique inchi keys
    are stored.
    """
    data = pd.DataFrame.from_dict({"inchi_key": inchi_keys})
    data = data.drop_duplicates("inchi_key")
    data.to_hdf(filename, "table")
    print(f"Created HDF5 stock with {len(data)} unique compounds")


def make_molbloom(
    smiles_list: _StrIterator, filename: str, filter_size: int, approx_mols: int
) -> None:
    """
    Put all the unique SMILES in a new bloom filter.

    :params smiles_list: the SMILES
    :params filename: the path to the saved filter
    :params filter_size: the size of the filter in bits
    :params approx_mols: approximately the number of compounds
    """
    filter_ = molbloom.CustomFilter(filter_size, approx_mols, "myfilter")
    processed_smiles = set()
    for smiles in smiles_list:
        try:
            smiles_can = Chem.CanonSmiles(smiles)
        # pylint: disable=broad-except
        except Exception:
            print(
                f"Failed to convert {smiles} to canonical SMILES.",
                flush=True,
            )
            continue
        if smiles_can in processed_smiles:
            continue
        filter_.add(smiles_can)
        processed_smiles.add(smiles_can)
    filter_.save(filename)
    print(f"Created bloom stock with {len(processed_smiles)} unique compounds")


def make_molbloom_inchi(
    inchi_keys: _StrIterator, filename: str, filter_size: int, approx_mols: int
) -> None:
    """
    Put all the unique InChI keys in a new bloom filter.

    :params inchi_keys: the Inchi Keys
    :params filename: the path to the saved filter
    :params filter_size: the size of the filter in bits
    :params approx_mols: approximately the number of compounds
    """
    filter_ = molbloom.CustomFilter(filter_size, approx_mols, "myfilter")
    nadded = 0
    for inchi_key in inchi_keys:
        filter_.add(inchi_key)
        nadded += 1
    filter_.save(filename)
    print(f"Created bloom stock with {nadded} unique compounds")


def make_mongo_stock(
    inchi_keys: _StrIterator, source_tag: str, host: Optional[str] = None
) -> None:
    """
    Put all the inchi keys from the given iterable in Mongo database as
    a molecules collection. Only unique inchi keys are stored.
    """
    mol_collection = MongoDbInchiKeyQuery(host=host).molecules
    if "inchi_key" not in mol_collection.index_information():
        mol_collection.create_index("inchi_key", name="inchi_key")
    mol_collection.delete_many({"source": source_tag})
    inchi_keys = set(inchi_keys)
    docs = ({"inchi_key": inchi_key, "source": source_tag} for inchi_key in inchi_keys)
    mol_collection.insert_many(docs)
    print(f"Created MongoDB stock with {len(inchi_keys)} unique compounds")


def main() -> None:
    """Entry-point for the smiles2stock tool"""
    args = _get_arguments()
    if args.source == "plain":
        smiles_gen = (smiles for smiles in extract_plain_smiles(args.files))
    else:
        smiles_gen = (smiles for smiles in extract_smiles_from_module(args.files))

    if not HAS_MOLBLOOM and args.target.startswith("molbloom"):
        raise ImportError(
            "Cannot create this stock format because it seems like molbloom is not installed. "
            "Please install aizynthfinder with extras dependencies."
        )

    if args.target == "molbloom":
        make_molbloom(smiles_gen, args.output, *args.bloom_params)
        return

    inchi_keys_gen = (inchi_key for inchi_key in _convert_smiles(smiles_gen))

    if args.target == "hdf5":
        make_hdf5_stock(inchi_keys_gen, args.output)
    elif args.target == "molbloom-inchi":
        make_molbloom_inchi(inchi_keys_gen, args.output, *args.bloom_params)
    else:
        make_mongo_stock(inchi_keys_gen, args.output, args.host)


if __name__ == "__main__":
    main()



""" Module containing classes to perform analysis of the tree search results.
"""

from __future__ import annotations

import operator
from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np
from paretoset import paretorank, paretoset

from aizynthfinder.analysis.utils import RouteSelectionArguments
from aizynthfinder.chem import FixedRetroReaction, hash_reactions
from aizynthfinder.context.scoring import StateScorer
from aizynthfinder.reactiontree import ReactionTree
from aizynthfinder.search.andor_trees import AndOrSearchTreeBase
from aizynthfinder.search.mcts import MctsNode, MctsSearchTree

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.context.scoring import Scorer
    from aizynthfinder.utils.type_utils import (
        Any,
        Dict,
        Iterable,
        List,
        Optional,
        Sequence,
        StrDict,
        Tuple,
        Union,
    )

    _Solution = Union[MctsNode, ReactionTree]
    _AnyListOfSolutions = Union[Sequence[MctsNode], Sequence[ReactionTree]]
    _AnyListOfReactions = Sequence[
        Union[Iterable[RetroReaction], Iterable[FixedRetroReaction]]
    ]


class TreeAnalysis:
    """
    Class that encapsulate various analysis that can be
    performed on a search tree.

    :ivar scorers: the objects used to score the nodes
    :ivar search_tree: the search tree

    :param search_tree: the search tree to do the analysis on
    :param scorer: the object used to score the nodes, defaults to StateScorer
    """

    def __init__(
        self,
        search_tree: Union[MctsSearchTree, AndOrSearchTreeBase],
        scorer: Optional[Union[Scorer, List[Scorer]]] = None,
    ) -> None:
        self.search_tree = search_tree
        scorer = scorer or StateScorer(search_tree.config)
        if isinstance(scorer, list):
            self.scorers: List[Scorer] = scorer
        else:
            self.scorers = [scorer]
        self._direction = "max"  # current implementation assumes maximisation
        self._single_objective = len(self.scorers) == 1

    def best(self) -> _Solution:
        """
        Returns the route or MCTS-like node with the highest score.
        If several routes have the same score, it will return the first

        :return: the top scoring node or route
        :raises ValueError: if this is a multi-objective analysis, a single best solution cannot be return
        """
        if not self._single_objective:
            raise ValueError("Cannot return best item for multi-objective analysis")

        if isinstance(self.search_tree, MctsSearchTree):
            nodes = self._all_nodes()
            sorted_nodes, _, _ = self.scorers[0].sort(nodes)
            return sorted_nodes[0]

        sorted_routes, _, _ = self.scorers[0].sort(self.search_tree.routes())
        return sorted_routes[0]

    def pareto_front(self) -> Tuple[_Solution, ...]:
        """
        Returns the Routes or MCTS-like Nodes that are the Pareto Front in the Multi-Objective search.

        :returns: the pareto front solutions.
        :raises ValueError: if this is a single-objective analysis, there is no Pareto-front
        """
        if self._single_objective:
            raise ValueError("Cannot return Pareto front for single-objective analysis")

        if isinstance(self.search_tree, MctsSearchTree):
            solutions = self.search_tree.nodes()
        else:
            solutions = self.search_tree.routes()  # type: ignore

        scores_arr = np.array(
            [[scorer(solution) for scorer in self.scorers] for solution in solutions]
        )
        direction_arr = np.repeat(self._direction, len(self.scorers))
        pareto_mask = paretoset(scores_arr, sense=direction_arr, distinct=False)
        pareto_idxs = np.arange(len(solutions))[pareto_mask]
        pareto_front: Sequence[_Solution] = [solutions[idx] for idx in pareto_idxs]

        return tuple(pareto_front)

    def sort(
        self, selection: Optional[RouteSelectionArguments] = None
    ) -> Tuple[_AnyListOfSolutions, Sequence[Dict[str, float]]]:
        """
        Sort and select the nodes or routes in the search tree.

        The score for each solution is a dictionary of scores,
        one for each of the objectives.

        :param selection: selection criteria for the routes
        :return: the sorted and selected items
        :return: the scores of the sorted items
        """
        selection = selection or RouteSelectionArguments()

        if not self._single_objective:
            return self._pareto_rank_sort(selection)

        scorer = self.scorers[0]
        if isinstance(self.search_tree, MctsSearchTree):
            nodes = self._all_nodes()
            sorted_items, sorted_scores, _ = scorer.sort(nodes)
            actions = [node.actions_to() for node in sorted_items]

        else:
            sorted_items, sorted_scores, _ = scorer.sort(self.search_tree.routes())
            actions = [route.reactions() for route in sorted_items]

        scores = [{repr(scorer): score} for score in sorted_scores]

        return self._collect_top_items(
            sorted_items, scores, actions, sorted_scores, selection, True
        )

    def tree_statistics(self) -> StrDict:
        """
        Returns statistics of the tree

        Currently it returns the number of nodes, the maximum number of transforms,
        maximum number of children, top score, if the top score route is solved,
        the number of molecule in the top score node, and information on pre-cursors

        :return: the statistics
        """
        if isinstance(self.search_tree, MctsSearchTree):
            return self._tree_statistics_mcts()
        return self._tree_statistics_andor()

    def _all_nodes(self) -> Sequence[MctsNode]:
        assert isinstance(self.search_tree, MctsSearchTree)
        # This is to keep backwards compatibility, this should be investigate further
        if repr(self.scorers[0]) == "state score":
            return list(self.search_tree.graph())
        return [node for node in self.search_tree.graph() if not node.children]

    def _pareto_rank_sort(
        self,
        selection: RouteSelectionArguments,
    ) -> Tuple[_AnyListOfSolutions, Sequence[Dict[str, float]]]:
        if isinstance(self.search_tree, MctsSearchTree):
            solutions = self._all_nodes()
        else:
            solutions = self.search_tree.routes()  # type: ignore

        scores_arr = np.array(
            [[scorer(solution) for scorer in self.scorers] for solution in solutions]
        )
        direction_arr = np.repeat(self._direction, len(self.scorers))
        pareto_ranks = paretorank(scores_arr, sense=direction_arr, distinct=False)

        sortidx = sorted(range(len(pareto_ranks)), key=pareto_ranks.__getitem__)
        sorted_pareto_ranks = sorted(pareto_ranks)
        sorted_scores = [
            {
                repr(scorer): scores_arr[idx, scorer_idx]
                for scorer_idx, scorer in enumerate(self.scorers)
            }
            for idx in sortidx
        ]
        sorted_items = [solutions[idx] for idx in sortidx]

        if isinstance(self.search_tree, MctsSearchTree):
            actions = [node.actions_to() for node in sorted_items]
        else:
            actions = [route.reactions() for route in sorted_items]  # type: ignore

        return self._collect_top_items(
            sorted_items, sorted_scores, actions, sorted_pareto_ranks, selection, False
        )

    def _top_nodes(self) -> Tuple[_Solution, ...]:
        if self._single_objective:
            return (self.best(),)
        return self.pareto_front()

    def _tree_statistics_andor(self) -> StrDict:
        assert isinstance(self.search_tree, AndOrSearchTreeBase)
        top_routes = self._top_nodes()
        mols_in_stock = self._top_ranked_join(
            ", ".join(mol.smiles for mol in route.leafs() if route.in_stock(mol))  # type: ignore
            for route in top_routes
        )
        mols_not_in_stock = self._top_ranked_join(
            ", ".join(mol.smiles for mol in route.leafs() if not route.in_stock(mol))  # type: ignore
            for route in top_routes
        )
        all_routes = self.search_tree.routes()
        policy_used_counts = self._policy_used_statistics(
            [reaction for route in all_routes for reaction in route.reactions()]
        )
        availability = self._top_ranked_join(
            ";".join(
                self.search_tree.config.stock.availability_string(mol)
                for mol in route.leafs()  # type: ignore
            )
            for route in top_routes
        )
        number_of_precursors_in_stock = self._top_ranked_join(
            sum(route.in_stock(leaf) for leaf in route.leafs()) for route in top_routes  # type: ignore
        )

        # For multi-objective it becomes to messy to have all the scores in this info,
        # they are nevertheless included with the routes
        if self._single_objective:
            top_score = self.scorers[0](top_routes[0])
        else:
            top_score = None

        return {
            "number_of_nodes": len(self.search_tree.mol_nodes),
            "max_transforms": max(
                node.prop["mol"].transform for node in self.search_tree.mol_nodes
            ),
            "max_children": max(
                len(node.children) for node in self.search_tree.mol_nodes
            ),
            "number_of_routes": len(all_routes),
            "number_of_solved_routes": sum(route.is_solved for route in all_routes),
            "top_score": top_score,
            "is_solved": self._top_ranked_join(route.is_solved for route in top_routes),
            "number_of_steps": self._top_ranked_join(
                len(list(route.reactions())) for route in top_routes  # type: ignore
            ),
            "number_of_precursors": self._top_ranked_join(
                len(list(route.leafs())) for route in top_routes  # type: ignore
            ),
            "number_of_precursors_in_stock": number_of_precursors_in_stock,
            "precursors_in_stock": mols_in_stock,
            "precursors_not_in_stock": mols_not_in_stock,
            "precursors_availability": availability,
            "policy_used_counts": policy_used_counts,
            "profiling": getattr(self.search_tree, "profiling", {}),
        }

    def _tree_statistics_mcts(self) -> StrDict:
        assert isinstance(self.search_tree, MctsSearchTree)
        top_nodes = self._top_nodes()
        assert isinstance(top_nodes[0], MctsNode)
        top_states = [node.state for node in top_nodes]  # type: ignore
        nodes = list(self.search_tree.graph())
        mols_in_stock = self._top_ranked_join(
            ", ".join(
                mol.smiles
                for mol, instock in zip(state.mols, state.in_stock_list)
                if instock
            )
            for state in top_states
        )
        mols_not_in_stock = self._top_ranked_join(
            ", ".join(
                mol.smiles
                for mol, instock in zip(state.mols, state.in_stock_list)
                if not instock
            )
            for state in top_states
        )

        policy_used_counts = self._policy_used_statistics(
            [node[child]["action"] for node in nodes for child in node.children]
        )

        # For multi-objective it becomes to messy to have all the scores in this info,
        # they are nevertheless included with the routes
        if self._single_objective:
            top_score = self.scorers[0](top_nodes[0])
        else:
            top_score = None

        return {
            "number_of_nodes": len(nodes),
            "max_transforms": max(node.state.max_transforms for node in nodes),
            "max_children": max(len(node.children) for node in nodes),
            "number_of_routes": sum(1 for node in nodes if not node.children),
            "number_of_solved_routes": sum(
                1 for node in nodes if not node.children and node.state.is_solved
            ),
            "top_score": top_score,
            "is_solved": self._top_ranked_join(state.is_solved for state in top_states),
            "number_of_steps": self._top_ranked_join(
                state.max_transforms for state in top_states
            ),
            "number_of_precursors": self._top_ranked_join(
                len(state.mols) for state in top_states
            ),
            "number_of_precursors_in_stock": self._top_ranked_join(
                sum(state.in_stock_list) for state in top_states
            ),
            "precursors_in_stock": mols_in_stock,
            "precursors_not_in_stock": mols_not_in_stock,
            "precursors_availability": self._top_ranked_join(
                ";".join(state.stock_availability) for state in top_states
            ),
            "policy_used_counts": policy_used_counts,
            "profiling": getattr(self.search_tree, "profiling", {}),
        }

    @staticmethod
    def _collect_top_items(
        items: _AnyListOfSolutions,
        scores: Sequence[Dict[str, float]],
        reactions: _AnyListOfReactions,
        ranks: Sequence[Union[int, float]],
        selection: RouteSelectionArguments,
        min_comp: bool,
    ) -> Tuple[_AnyListOfSolutions, Sequence[Dict[str, float]]]:
        """
        Finds and returns the top-ranked super-nodes or reaction trees and
        their scores.

        The `selection` argument is used to control the number of items that are returned.
        If at least one of the items is solved, and `selection` specifies that all items
        should be returned, all solved items are returned.
        Otherwise, at least a minimum number of items are returned, but there could be more
        than this if they have the same rank. However not more than a maximum number of
        items are returned.

        Duplicated routes are ignored.

        The items are compared based on rank. For single-ojective analysis, this is simply
        the score and higher is better. For multi-objective analysis, the rank is the Pareto
        rank and smaller i better.

        :param items: the super-nodes or reaction trees to select from, sorted by rank
        :param scores: the computed scores of the nodes
        :param reaction: the reactions for each item, used for checking duplicity
        :param ranks: the rank of each item, used for determining selection
        :param selection: min and max solutions, or indication to return all
        :param min_comp: True if higher rank is better, else False
        """
        if len(items) <= selection.nmin:
            return items, scores

        max_return, min_return = selection.nmax, selection.nmin
        if selection.return_all:
            nsolved = sum(int(item.is_solved) for item in items)
            if nsolved:
                max_return = nsolved
                min_return = nsolved

        seen_hashes = set()
        best_items: List[Any] = []
        best_scores = []
        if min_comp:
            last_rank = 1e16
            comp_op = operator.lt
        else:
            last_rank = 0
            comp_op = operator.gt
        for rank, score, item, actions in zip(ranks, scores, items, reactions):
            if len(best_items) >= min_return and comp_op(rank, last_rank):
                break
            route_hash = hash_reactions(actions)

            if route_hash in seen_hashes:
                continue
            seen_hashes.add(route_hash)
            best_items.append(item)
            best_scores.append(score)
            last_rank = rank

            if max_return and len(best_items) == max_return:
                break

        return best_items, best_scores

    @staticmethod
    def _policy_used_statistics(
        reactions: Iterable[Union[RetroReaction, FixedRetroReaction]]
    ) -> StrDict:
        policy_used_counts: StrDict = defaultdict(lambda: 0)
        for reaction in reactions:
            policy_used = reaction.metadata.get("policy_name")
            if policy_used:
                policy_used_counts[policy_used] += 1
        return dict(policy_used_counts)

    @staticmethod
    def _top_ranked_join(items: Iterable[Any]) -> Union[str, Any]:
        items = list(items)
        if len(items) == 1:
            return items[0]
        return "|".join(f"{item}" for item in items)



""" Module containing classes that implements different expansion policy strategies
"""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from aizynthfinder.chem import SmilesBasedRetroReaction, TemplatedRetroReaction
from aizynthfinder.context.policy.utils import _make_fingerprint
from aizynthfinder.utils.exceptions import PolicyException
from aizynthfinder.utils.logging import logger
from aizynthfinder.utils.models import load_model

if TYPE_CHECKING:
    from aizynthfinder.chem import TreeMolecule
    from aizynthfinder.chem.reaction import RetroReaction
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.utils.type_utils import (
        Any,
        Dict,
        List,
        Optional,
        Sequence,
        StrDict,
        Tuple,
    )


class ExpansionStrategy(abc.ABC):
    """
    A base class for all expansion strategies.

    The strategy can be used by either calling the `get_actions` method
    of by calling the instantiated class with a list of molecule.

    .. code-block::

        expander = MyExpansionStrategy("dummy", config)
        actions, priors = expander.get_actions(molecules)
        actions, priors = expander(molecules)

    :param key: the key or label
    :param config: the configuration of the tree search
    """

    _required_kwargs: List[str] = []

    def __init__(self, key: str, config: Configuration, **kwargs: str) -> None:
        if any(name not in kwargs for name in self._required_kwargs):
            raise PolicyException(
                f"A {self.__class__.__name__} class needs to be initiated "
                f"with keyword arguments: {', '.join(self._required_kwargs)}"
            )
        self._config = config
        self._logger = logger()
        self.key = key

    def __call__(
        self,
        molecules: Sequence[TreeMolecule],
        cache_molecules: Optional[Sequence[TreeMolecule]] = None,
    ) -> Tuple[List[RetroReaction], List[float]]:
        return self.get_actions(molecules, cache_molecules)

    @abc.abstractmethod
    def get_actions(
        self,
        molecules: Sequence[TreeMolecule],
        cache_molecules: Optional[Sequence[TreeMolecule]] = None,
    ) -> Tuple[List[RetroReaction], List[float]]:
        """
        Get all the probable actions of a set of molecules

        :param molecules: the molecules to consider
        :param cache_molecules: additional molecules to submit to the expansion
                                  policy but that only will be cached for later use
        :return: the actions and the priors of those actions
        """

    def reset_cache(self) -> None:
        """Reset the prediction cache"""


class MultiExpansionStrategy(ExpansionStrategy):
    """
    A base class for combining multiple expansion strategies.

    The strategy can be used by either calling the `get_actions` method
    or by calling the instantiated class with a list of molecules.

    :ivar expansion_strategy_keys: the keys of the selected expansion strategies
    :ivar additive_expansion: a conditional setting to specify whether all the actions
        and priors of the selected expansion strategies should be combined or not.
        Defaults to False.
    :ivar expansion_strategy_weights: a list of weights for each expansion strategy.
        The weights should sum to one. Exception is the default, where unity weight
        is associated to each strategy.

    :param key: the key or label
    :param config: the configuration of the tree search
    :param expansion_strategies: the keys of the selected expansion strategies. All keys
        of the selected expansion strategies must exist in the expansion policies listed
        in config
    """

    _required_kwargs = ["expansion_strategies"]

    def __init__(
        self,
        key: str,
        config: Configuration,
        **kwargs: Any,
    ) -> None:
        super().__init__(key, config, **kwargs)
        self._config = config
        self._expansion_strategies: List[ExpansionStrategy] = []
        self.expansion_strategy_keys = kwargs["expansion_strategies"]

        self.cutoff_number = kwargs.get("cutoff_number")
        if self.cutoff_number:
            print(f"Setting multi-expansion cutoff_number: {self.cutoff_number}")

        self.expansion_strategy_weights = self._set_expansion_strategy_weights(kwargs)
        self.additive_expansion: bool = bool(kwargs.get("additive_expansion", False))
        self._logger.info(
            f"Multi-expansion strategy with policies: {self.expansion_strategy_keys}"
            f", and corresponding weights: {self.expansion_strategy_weights}"
        )

    def get_actions(
        self,
        molecules: Sequence[TreeMolecule],
        cache_molecules: Optional[Sequence[TreeMolecule]] = None,
    ) -> Tuple[List[RetroReaction], List[float]]:
        """
        Get all the probable actions of a set of molecules, using the selected policies.

        The default implementation combines all the actions and priors of the
        selected expansion strategies into two lists respectively if the
        'additive_expansion' setting is set to True. This function can be overridden by
        a sub class to combine different expansion strategies in different ways.

        :param molecules: the molecules to consider
        :param cache_molecules: additional molecules to submit to the expansion
            policy but that only will be cached for later use
        :return: the actions and the priors of those actions
        :raises: PolicyException: if the policy isn't selected
        """
        expansion_strategies = self._get_expansion_strategies_from_config()

        all_possible_actions = []
        all_priors = []
        for expansion_strategy, expansion_strategy_weight in zip(
            expansion_strategies, self.expansion_strategy_weights
        ):
            possible_actions, priors = expansion_strategy.get_actions(
                molecules, cache_molecules
            )

            all_possible_actions.extend(possible_actions)
            if not self.additive_expansion and all_possible_actions:
                all_priors.extend(priors)
                break

            weighted_prior = [expansion_strategy_weight * p for p in priors]

            all_priors.extend(weighted_prior)

        all_possible_actions, all_priors = self._prune_actions(
            all_possible_actions, all_priors
        )
        return all_possible_actions, all_priors

    def _get_expansion_strategies_from_config(self) -> List[ExpansionStrategy]:
        if self._expansion_strategies:
            return self._expansion_strategies

        if not all(
            key in self._config.expansion_policy.items
            for key in self.expansion_strategy_keys
        ):
            raise ValueError(
                "The input expansion strategy keys must exist in the "
                "expansion policies listed in config"
            )
        self._expansion_strategies = [
            self._config.expansion_policy[key] for key in self.expansion_strategy_keys
        ]

        for expansion_strategy, weight in zip(
            self._expansion_strategies, self.expansion_strategy_weights
        ):
            if not getattr(expansion_strategy, "rescale_prior", True) and weight < 1:
                setattr(expansion_strategy, "rescale_prior", True)
                self._logger.info(
                    f"Enforcing {expansion_strategy.key}.rescale_prior=True"
                )
        return self._expansion_strategies

    def _prune_actions(
        self, actions: List[RetroReaction], priors: List[float]
    ) -> Tuple[List[RetroReaction], List[float]]:
        """
        Prune the actions if a maximum number of actions is specified.

        :param actions: list of predicted actions
        :param priors: list of prediction probabilities
        :return: the top 'self.cutoff_number' actions and corresponding priors.
        """
        if not self.cutoff_number:
            return actions, priors

        sortidx = np.argsort(np.array(priors))[::-1].astype(int)
        priors = [priors[idx] for idx in sortidx[0 : self.cutoff_number]]
        actions = [actions[idx] for idx in sortidx[0 : self.cutoff_number]]
        return actions, priors

    def _set_expansion_strategy_weights(self, kwargs: StrDict) -> List[float]:
        """
        Set the weights of each expansion strategy using the input kwargs from config.
        The weights in the config should sum to one.
        If not set in the config file, the weights default to one for each strategy
        (for backwards compatibility).

        :param kwargs: input arguments to the MultiExpansionStrategy
        :raises: ValueError if weights from the config file do not sum to one.
        :return: a list of expansion strategy weights
        """
        if not "expansion_strategy_weights" in kwargs:
            return [1.0 for _ in self.expansion_strategy_keys]

        expansion_strategy_weights = kwargs["expansion_strategy_weights"]
        sum_weights = sum(expansion_strategy_weights)

        if sum_weights != 1:
            raise ValueError(
                "The expansion strategy weights in MultiExpansion should "
                "sum to one. -> "
                f"sum({expansion_strategy_weights})={sum_weights}."
            )

        return expansion_strategy_weights


class TemplateBasedExpansionStrategy(ExpansionStrategy):
    """
    A template-based expansion strategy that will return `TemplatedRetroReaction` objects upon expansion.

    :ivar template_column: the column in the template file that contains the templates
    :ivar cutoff_cumulative: the accumulative probability of the suggested templates
    :ivar cutoff_number: the maximum number of templates to returned
    :ivar use_rdchiral: a boolean to apply templates with RDChiral
    :ivar use_remote_models: a boolean to connect to remote TensorFlow servers
    :ivar rescale_prior: a boolean to apply softmax to the priors
    :ivar chiral_fingerprints: if True will base expansion on chiral fingerprint
    :ivar mask: a boolean vector of masks for the reaction templates. The length of the vector should be equal to the
        number of templates. It is set to None if no mask file is provided as input.

    :param key: the key or label
    :param config: the configuration of the tree search
    :param model: the source of the policy model
    :param template: the path to a HDF5 file with the templates
    :raises PolicyException: if the length of the model output vector is not same as the
        number of templates
    """

    _required_kwargs = [
        "model",
        "template",
    ]

    def __init__(self, key: str, config: Configuration, **kwargs: str) -> None:
        super().__init__(key, config, **kwargs)

        source = kwargs["model"]
        templatefile = kwargs["template"]
        maskfile: str = kwargs.get("mask", "")
        self.template_column: str = kwargs.get("template_column", "retro_template")
        self.cutoff_cumulative: float = float(kwargs.get("cutoff_cumulative", 0.995))
        self.cutoff_number: int = int(kwargs.get("cutoff_number", 50))
        self.use_rdchiral: bool = bool(kwargs.get("use_rdchiral", True))
        self.use_remote_models: bool = bool(kwargs.get("use_remote_models", False))
        self.rescale_prior: bool = bool(kwargs.get("rescale_prior", False))
        self.chiral_fingerprints = bool(kwargs.get("chiral_fingerprints", False))

        self._logger.info(
            f"Loading template-based expansion policy model from {source} to {self.key}"
        )
        self.model = load_model(source, self.key, self.use_remote_models)

        self._logger.info(f"Loading templates from {templatefile} to {self.key}")
        if templatefile.endswith(".csv.gz") or templatefile.endswith(".csv"):
            self.templates: pd.DataFrame = pd.read_csv(
                templatefile, index_col=0, sep="\t"
            )
        else:
            self.templates = pd.read_hdf(templatefile, "table")

        self.mask: Optional[np.ndarray] = (
            self._load_mask_file(maskfile) if maskfile else None
        )

        if hasattr(self.model, "output_size") and len(self.templates) != self.model.output_size:  # type: ignore
            raise PolicyException(
                f"The number of templates ({len(self.templates)}) does not agree with the "  # type: ignore
                f"output dimensions of the model ({self.model.output_size})"
            )
        self._cache: Dict[str, Tuple[np.ndarray, np.ndarray]] = {}

    def get_actions(
        self,
        molecules: Sequence[TreeMolecule],
        cache_molecules: Optional[Sequence[TreeMolecule]] = None,
    ) -> Tuple[List[RetroReaction], List[float]]:
        """
        Get all the probable actions of a set of molecules, using the selected policies and given cutoffs

        :param molecules: the molecules to consider
        :param cache_molecules: additional molecules to submit to the expansion
                                  policy but that only will be cached for later use
        :return: the actions and the priors of those actions
        """

        possible_actions = []
        priors: List[float] = []
        cache_molecules = cache_molecules or []
        self._update_cache(list(molecules) + list(cache_molecules))

        for mol in molecules:
            probable_transforms_idx, probs = self._cache[mol.inchi_key]
            possible_moves = self.templates.iloc[probable_transforms_idx]
            if self.rescale_prior:
                probs /= probs.sum()
            priors.extend(probs)
            for idx, (move_index, move) in enumerate(possible_moves.iterrows()):
                metadata = dict(move)
                del metadata[self.template_column]
                metadata["policy_probability"] = float(probs[idx].round(4))
                metadata["policy_probability_rank"] = idx
                metadata["policy_name"] = self.key
                metadata["template_code"] = move_index
                metadata["template"] = move[self.template_column]
                possible_actions.append(
                    TemplatedRetroReaction(
                        mol,
                        smarts=move[self.template_column],
                        metadata=metadata,
                        use_rdchiral=self.use_rdchiral,
                    )
                )
        return possible_actions, priors  # type: ignore

    def reset_cache(self) -> None:
        """Reset the prediction cache"""
        self._cache = {}

    def _cutoff_predictions(self, predictions: np.ndarray) -> np.ndarray:
        """
        Get the top transformations, by selecting those that have:
            * cumulative probability less than a threshold (cutoff_cumulative)
            * or at most N (cutoff_number)
        """
        if self.mask is not None:
            predictions[~self.mask] = 0
        sortidx = np.argsort(predictions)[::-1]
        cumsum: np.ndarray = np.cumsum(predictions[sortidx])
        if any(cumsum >= self.cutoff_cumulative):
            maxidx = int(np.argmin(cumsum < self.cutoff_cumulative))
        else:
            maxidx = len(cumsum)
        maxidx = min(maxidx, self.cutoff_number) or 1
        return sortidx[:maxidx]

    def _load_mask_file(self, maskfile: str) -> np.ndarray:
        self._logger.info(f"Loading masking of templates from {maskfile} to {self.key}")
        mask = np.load(maskfile)["arr_0"]
        if len(mask) != len(self.templates):
            raise PolicyException(
                f"The number of masks {len(mask)} does not match the number of templates {len(self.templates)}"
            )
        return mask

    def _update_cache(self, molecules: Sequence[TreeMolecule]) -> None:
        pred_inchis = []
        fp_list = []
        for molecule in molecules:
            if molecule.inchi_key in self._cache or molecule.inchi_key in pred_inchis:
                continue
            fp_list.append(
                _make_fingerprint(molecule, self.model, self.chiral_fingerprints)
            )
            pred_inchis.append(molecule.inchi_key)

        if not pred_inchis:
            return

        pred_list = np.asarray(self.model.predict(np.vstack(fp_list)))
        for pred, inchi in zip(pred_list, pred_inchis):
            probable_transforms_idx = self._cutoff_predictions(pred)
            self._cache[inchi] = (
                probable_transforms_idx,
                pred[probable_transforms_idx],
            )


class TemplateBasedDirectExpansionStrategy(TemplateBasedExpansionStrategy):
    """
    A template-based expansion strategy that will return `SmilesBasedRetroReaction` objects upon expansion
    by directly applying the template

    :param key: the key or label
    :param config: the configuration of the tree search
    :param source: the source of the policy model
    :param templatefile: the path to a HDF5 file with the templates
    :raises PolicyException: if the length of the model output vector is not same as the number of templates
    """

    def get_actions(
        self,
        molecules: Sequence[TreeMolecule],
        cache_molecules: Optional[Sequence[TreeMolecule]] = None,
    ) -> Tuple[List[RetroReaction], List[float]]:
        """
        Get all the probable actions of a set of molecules, using the selected policies and given cutoffs

        :param molecules: the molecules to consider
        :param cache_molecules: additional molecules to submit to the expansion
            policy but that only will be cached for later use
        :return: the actions and the priors of those actions
        """
        possible_actions = []
        priors = []

        super_actions, super_priors = super().get_actions(molecules, cache_molecules)
        for templated_action, prior in zip(super_actions, super_priors):
            for reactants in templated_action.reactants:
                reactants_str = ".".join(mol.smiles for mol in reactants)
                new_action = SmilesBasedRetroReaction(
                    templated_action.mol,
                    metadata=templated_action.metadata,
                    reactants_str=reactants_str,
                )
                possible_actions.append(new_action)
                priors.append(prior)

        return possible_actions, priors  # type: ignore



""" Module containing a classes representation various tree nodes
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from aizynthfinder.chem import TreeMolecule
from aizynthfinder.search.andor_trees import TreeNodeMixin

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.context.config import Configuration
    from aizynthfinder.search.dfpn import SearchTree
    from aizynthfinder.utils.type_utils import List, Optional, Sequence, Set, StrDict

BIG_INT = int(1e10)


class _SuperNode(TreeNodeMixin):
    def __init__(self) -> None:
        # pylint: disable=invalid-name
        self.pn = 1  # Proof-number
        self.dn = 1  # Disproof-number
        self.pn_threshold = BIG_INT
        self.dn_threshold = BIG_INT
        self._children: List["_SuperNode"] = []
        self.expandable = True

    @property  # type: ignore
    def children(self) -> List[ReactionNode]:  # type: ignore
        """Gives the reaction children nodes"""
        return self._children  # type: ignore

    @property
    def closed(self) -> bool:
        """Return if the node is proven or disproven"""
        return self.proven or self.disproven

    @property
    def proven(self) -> bool:
        """Return if the node is proven"""
        return self.pn == 0

    @property
    def disproven(self) -> bool:
        """Return if the node is disproven"""
        return self.dn == 0

    def explorable(self) -> bool:
        """Return if the node can be explored by the search algorithm"""
        return not (
            self.closed or self.pn > self.pn_threshold or self.dn > self.dn_threshold
        )

    def reset(self) -> None:
        """Reset the thresholds"""
        if self.closed or self.expandable:
            return
        for child in self._children:
            child.reset()
        self.update()
        self.pn_threshold = BIG_INT
        self.dn_threshold = BIG_INT

    def update(self) -> None:
        """Update the proof and disproof numbers"""
        raise NotImplementedError("Implement a child class")

    def _set_disproven(self) -> None:
        self.pn = BIG_INT
        self.dn = 0

    def _set_proven(self) -> None:
        self.pn = 0
        self.dn = BIG_INT


class MoleculeNode(_SuperNode):
    """
    An OR node representing a molecule

    :ivar expandable: if True, this node is part of the frontier
    :ivar mol: the molecule represented by the node
    :ivar in_stock: if True the molecule is in stock and hence should not be expanded
    :ivar parent: the parent of the node
    :ivar pn: the proof number
    :ivar dn: the disproof number
    :ivar pn_threshold: the threshold for proof number
    :ivar dn_threshold: the threshold for disproof number

    :param mol: the molecule to be represented by the node
    :param config: the configuration of the search
    :param parent: the parent of the node, optional
    """

    def __init__(
        self,
        mol: TreeMolecule,
        config: Configuration,
        owner: SearchTree,
        parent: Optional[ReactionNode] = None,
    ) -> None:
        super().__init__()

        self.mol = mol
        self._config = config
        self.in_stock = mol in config.stock
        self.parent = parent
        self._edge_costs: List[int] = []
        self.tree = owner

        # Makes it unexpandable if we have reached maximum depth
        self.expandable = self.mol.transform < self._config.search.max_transforms

        if self.in_stock:
            self.expandable = False
            self._set_proven()
        elif not self.expandable:
            self._set_disproven()

    @classmethod
    def create_root(
        cls, smiles: str, config: Configuration, owner: SearchTree
    ) -> "MoleculeNode":
        """
        Create a root node for a tree using a SMILES.

        :param smiles: the SMILES representation of the root state
        :param config: settings of the tree search algorithm
        :return: the created node
        """
        mol = TreeMolecule(parent=None, transform=0, smiles=smiles)
        return MoleculeNode(mol=mol, config=config, owner=owner)

    @property
    def prop(self) -> StrDict:
        return {"solved": self.proven, "mol": self.mol}

    def expand(self) -> None:
        """Expand the molecule by utilising an expansion policy"""
        self.expandable = False
        reactions, priors = self._config.expansion_policy([self.mol])
        self.tree.profiling["expansion_calls"] += 1

        if not reactions:
            self._set_disproven()
            return

        costs = -np.log(np.clip(priors, 1e-3, 1.0))
        reaction_costs = []
        reactions_to_expand = []
        for reaction, cost in zip(reactions, costs):
            try:
                _ = reaction.reactants
                self.tree.profiling["reactants_generations"] += 1
            except:  # pylint: disable=bare-except
                continue
            if not reaction.reactants:
                continue
            for idx, _ in enumerate(reaction.reactants):
                rxn_copy = reaction.copy(idx)
                reactions_to_expand.append(rxn_copy)
                reaction_costs.append(cost)

        for cost, rxn in zip(reaction_costs, reactions_to_expand):
            self._add_child(rxn, cost)

        if not self._children:
            self._set_disproven()

    def promising_child(self) -> Optional[ReactionNode]:
        """
        Find and return the most promising child for exploration
        Updates the thresholds on that child
        """
        min_indices = np.argsort(
            [
                edge_cost + child.pn if not child.closed else BIG_INT
                for edge_cost, child in zip(self._edge_costs, self._children)
            ]
        )
        best_child = self._children[min_indices[0]]
        if len(self._children) > 1 and not self._children[min_indices[1]].closed:
            s2_pn = self._children[min_indices[1]].pn
        else:
            s2_pn = BIG_INT

        best_child.pn_threshold = (
            min(self.pn_threshold, s2_pn + 2) - self._edge_costs[min_indices[0]]
        )
        best_child.dn_threshold = self.dn_threshold - self.dn + best_child.dn
        return best_child

    def update(self) -> None:
        """Update the proof and disproof numbers"""
        func = all if self.parent is None else any
        if func(child.proven for child in self._children):
            self._set_proven()
            return
        if all(child.disproven for child in self._children):
            self._set_disproven()
            return

        child_dns = [child.dn for child in self._children if not child.closed]
        if not child_dns:
            self._set_proven()
            return

        self.dn = sum(child_dns)
        if self.dn >= BIG_INT:
            self.pn = 0
        else:
            self.pn = min(
                edge_cost + child.pn
                for edge_cost, child in zip(self._edge_costs, self._children)
                if not child.closed
            )
        return

    def _add_child(self, reaction: RetroReaction, _: float) -> None:
        reactants = reaction.reactants[reaction.index]
        if not reactants:
            return

        ancestors = self._ancestors()
        for mol in reactants:
            if mol in ancestors:
                return

        rxn_node = ReactionNode(
            reaction=reaction, config=self._config, owner=self.tree, parent=self
        )
        self._children.append(rxn_node)
        self._edge_costs.append(1)

    def _ancestors(self) -> Set[TreeMolecule]:
        if not self.parent:
            return {self.mol}

        # pylint: disable=protected-access
        ancestors = self.parent.parent._ancestors()
        ancestors.add(self.mol)
        return ancestors


class ReactionNode(_SuperNode):
    """
    An AND node representing a reaction

    :ivar parent: the parent of the node
    :ivar reaction: the reaction represented by the node
    :ivar pn: the proof number
    :ivar dn: the disproof number
    :ivar pn_threshold: the threshold for proof number
    :ivar dn_threshold: the threshold for disproof number
    :ivar expandable: if the node is expandable

    :param reaction: the reaction to be represented by the node
    :param config: the configuration of the search
    :param parent: the parent of the node
    """

    def __init__(
        self,
        reaction: RetroReaction,
        config: Configuration,
        owner: SearchTree,
        parent: MoleculeNode,
    ) -> None:
        super().__init__()
        self._config = config
        self.parent = parent
        self.reaction = reaction
        self.tree = owner

    @property  # type: ignore
    def children(self) -> List[MoleculeNode]:  # type: ignore
        """Gives the molecule children nodes"""
        return self._children  # type: ignore

    @property
    def prop(self) -> StrDict:
        return {"solved": self.proven, "reaction": self.reaction}

    @property
    def proven(self) -> bool:
        """Return if the node is proven"""
        if self.expandable:
            return False
        if self.pn == 0:
            return True
        return all(child.proven for child in self._children)

    @property
    def disproven(self) -> bool:
        """Return if the node is disproven"""
        if self.expandable:
            return False
        if self.dn == 0:
            return True
        return any(child.disproven for child in self._children)

    def expand(self) -> None:
        """Expand the node by creating nodes for each reactant"""
        self.expandable = False
        reactants = self.reaction.reactants[self.reaction.index]
        self._children = [
            MoleculeNode(mol=mol, config=self._config, owner=self.tree, parent=self)
            for mol in reactants
        ]

    def promising_child(self) -> Optional[MoleculeNode]:
        """
        Find and return the most promising child for exploration
        Updates the thresholds on that child
        """
        min_indices = np.argsort(
            [child.dn if not child.closed else BIG_INT for child in self._children]
        )

        best_child = self._children[min_indices[0]]
        if len(self._children) > 1 and not self._children[min_indices[1]].closed:
            s2_dn = self._children[min_indices[1]].dn
        else:
            s2_dn = BIG_INT

        best_child.pn_threshold = self.pn_threshold - self.pn + best_child.pn
        best_child.dn_threshold = min(self.dn_threshold, s2_dn + 1)
        return best_child

    def update(self) -> None:
        """Update the proof and disproof numbers"""
        if all(child.proven for child in self._children):
            self._set_proven()
            return
        if any(child.disproven for child in self._children):
            self._set_disproven()
            return

        self.pn = sum(child.pn for child in self._children if not child.closed)
        self.dn = min(child.dn for child in self._children if not child.closed)



"""Module containing routines to work with files and processes."""
from __future__ import annotations

import gzip
import json
import subprocess
import tempfile
import time
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
from deprecated import deprecated

from aizynthfinder.utils.logging import logger

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import (
        Any,
        Callable,
        List,
        Optional,
        Sequence,
        Union,
    )


def read_datafile(filename: Union[str, Path]) -> pd.DataFrame:
    """
    Read aizynth output from disc in either .hdf5 or .json format

    :param filename: the path to the data
    :return: the loaded data
    """
    filename_str = str(filename)
    if filename_str.endswith(".hdf5") or filename_str.endswith(".hdf"):
        return pd.read_hdf(filename, "table")
    return pd.read_json(filename, orient="table")


def save_datafile(data: pd.DataFrame, filename: Union[str, Path]) -> None:
    """
    Save the given data to disc in either .hdf5 or .json format

    :param data: the data to save
    :param filename: the path to the data
    """
    filename_str = str(filename)
    if filename_str.endswith(".hdf5") or filename_str.endswith(".hdf"):
        with warnings.catch_warnings():  # This wil suppress a PerformanceWarning
            warnings.simplefilter("ignore")
            data.to_hdf(filename, key="table")
    else:
        data.to_json(filename, orient="table")


@deprecated(version="4.0.0", reason="replaced by 'cat_datafiles'")
def cat_hdf_files(
    input_files: List[str], output_name: str, trees_name: Optional[str] = None
) -> None:
    """Concatenate hdf5 or json datafiles"""
    cat_datafiles(input_files, output_name, trees_name)


def cat_datafiles(
    input_files: List[str], output_name: str, trees_name: Optional[str] = None
) -> None:
    """
    Concatenate hdf5 or json datafiles

    if `tree_name` is given, will take out the `trees` column
    from the tables and save it to a gzipped-json file.

    :param input_files: the paths to the files to concatenate
    :param output_name: the name of the concatenated file
    :param trees_name: the name of the concatenated trees
    """
    data = read_datafile(input_files[0])
    if "trees" not in data.columns:
        trees_name = None

    if trees_name:
        columns = list(data.columns)
        columns.remove("trees")
        trees = list(data["trees"].values)
        data = data[columns]

    for filename in input_files[1:]:
        new_data = read_datafile(filename)
        if trees_name:
            trees.extend(new_data["trees"].values)
            new_data = new_data[columns]
        data = pd.concat([data, new_data])

    save_datafile(data.reset_index(drop=True), output_name)
    if trees_name:
        if not trees_name.endswith(".gz"):
            trees_name += ".gz"
        with gzip.open(trees_name, "wt", encoding="UTF-8") as fileobj:
            json.dump(trees, fileobj)


def split_file(filename: str, nparts: int) -> List[str]:
    """
    Split the content of a text file into a given number of temporary files

    :param filename: the path to the file to split
    :param nparts: the number of parts to create
    :return: list of filenames of the parts
    """
    with open(filename, "r") as fileobj:
        lines = fileobj.read().splitlines()

    filenames = []
    batch_size, remainder = divmod(len(lines), nparts)
    stop = 0
    for part in range(1, nparts + 1):
        start = stop
        stop += batch_size + 1 if part <= remainder else batch_size
        filenames.append(tempfile.mktemp())
        with open(filenames[-1], "w") as fileobj:
            fileobj.write("\n".join(lines[start:stop]))
    return filenames


def start_processes(
    inputs: Sequence[Any], log_prefix: str, cmd_callback: Callable, poll_freq: int = 5
) -> None:
    """
    Start a number of background processes and wait for them
    to complete.

    The standard output and standard error is saved to a log file.

    The command to start for each process is given by the ``cmd_callback``
    function that takes the index of the process and an item of the input
    as arguments.

    :param inputs: a sequence of input to the processes
    :param log_prefix: the prefix to the log file of each processes
    :param cmd_callback: function that creates the process commands
    :param poll_freq: the polling frequency for checking if processes are completed
    """
    processes = []
    output_fileobjs = []
    for index, iinput in enumerate(inputs, 1):
        output_fileobjs.append(open(f"{log_prefix}{index}.log", "w"))
        cmd = cmd_callback(index, iinput)
        processes.append(
            subprocess.Popen(cmd, stdout=output_fileobjs[-1], stderr=subprocess.STDOUT)
        )
        logger().info(f"Started background task with pid={processes[-1].pid}")

    logger().info("Waiting for background tasks to complete...")
    not_finished = True
    while not_finished:
        time.sleep(poll_freq)
        not_finished = False
        for process, fileobj in zip(processes, output_fileobjs):
            fileobj.flush()
            if process.poll() is None:
                not_finished = True

    for fileobj in output_fileobjs:
        fileobj.close()



""" Module containing classes to deal with Molecules - mostly wrappers around rdkit routines.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem, Descriptors

from aizynthfinder.utils.bonds import sort_bonds
from aizynthfinder.utils.exceptions import MoleculeException

if TYPE_CHECKING:
    from aizynthfinder.utils.type_utils import (
        Callable,
        Dict,
        List,
        Optional,
        RdMol,
        Sequence,
        Tuple,
        Union,
    )


class Molecule:
    """
    A base class for molecules. Encapsulate an RDKit mol object and
    functions that can be applied to such a molecule.

    The objects of this class is hashable by the inchi key and hence
    comparable with the equality operator.

    :ivar rd_mol: the RDkit mol object that is encapsulated
    :ivar smiles: the SMILES representation of the molecule

    :param rd_mol: a RDKit mol object to encapsulate, defaults to None
    :param smiles: a SMILES to convert to a molecule object, defaults to None
    :param sanitize: if True, the molecule will be immediately sanitized, defaults to False
    :raises MoleculeException: if neither rd_mol or smiles is given, or if the molecule could not be sanitized
    """

    def __init__(
        self,
        rd_mol: Optional[RdMol] = None,
        smiles: Optional[str] = None,
        sanitize: bool = False,
    ) -> None:
        if not rd_mol and not smiles:
            raise MoleculeException(
                "Need to provide either a rdkit Mol object or smiles to create a molecule"
            )

        if rd_mol:
            self.rd_mol = rd_mol
            self.smiles = Chem.MolToSmiles(rd_mol)
        else:
            self.smiles = smiles
            self.rd_mol = Chem.MolFromSmiles(smiles, sanitize=False)

        self._inchi_key: Optional[str] = None
        self._inchi: Optional[str] = None
        self._fingerprints: Dict[Union[Tuple[int, int], Tuple[int]], np.ndarray] = {}
        self._is_sanitized: bool = False

        # Atom mapping -> atom index dictionary
        self._atom_mappings: Dict[int, int] = {}
        # Atom index -> atom mapping dictionary
        self._reverse_atom_mappings: Dict[int, int] = {}

        if sanitize:
            self.sanitize()

    def __hash__(self) -> int:
        return hash(self.inchi_key)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Molecule):
            return False
        return self.inchi_key == other.inchi_key

    def __len__(self) -> int:
        return self.rd_mol.GetNumAtoms()

    def __str__(self) -> str:
        return self.smiles

    @property
    def inchi(self) -> str:
        """
        The inchi representation of the molecule
        Created by lazy evaluation. Will cause the molecule to be sanitized.

        :return: the inchi
        """
        if not self._inchi:
            self.sanitize(raise_exception=False)
            self._inchi = Chem.MolToInchi(self.rd_mol)
            if self._inchi is None:
                raise MoleculeException("Could not make InChI")
        return self._inchi

    @property
    def inchi_key(self) -> str:
        """
        The inchi key representation of the molecule
        Created by lazy evaluation. Will cause the molecule to be sanitized.

        :return: the inchi key
        """
        if not self._inchi_key:
            self.sanitize(raise_exception=False)
            self._inchi_key = Chem.MolToInchiKey(self.rd_mol)
            if self._inchi_key is None:
                raise MoleculeException("Could not make InChI key")
        return self._inchi_key

    @property
    def index_to_mapping(self) -> Dict[int, int]:
        """Return a dictionary mapping to atom indices to atom mappings"""
        if not self._reverse_atom_mappings:
            self._reverse_atom_mappings = {
                index: mapping for mapping, index in self.mapping_to_index.items()
            }
        return self._reverse_atom_mappings

    @property
    def mapping_to_index(self) -> Dict[int, int]:
        """Return a dictionary mapping to atom mappings to atom indices"""
        if not self._atom_mappings:
            self._atom_mappings = {
                atom.GetAtomMapNum(): atom.GetIdx()
                for atom in self.rd_mol.GetAtoms()
                if atom.GetAtomMapNum()
            }
        return self._atom_mappings

    @property
    def weight(self) -> float:
        """Return the exact molecular weight of the molecule"""
        self.sanitize(raise_exception=False)
        return Descriptors.ExactMolWt(self.rd_mol)

    def basic_compare(self, other: "Molecule") -> bool:
        """
        Compare this molecule to another but only to
        the basic part of the inchi key, thereby ignoring stereochemistry etc

        :param other: the molecule to compare to
        :return: True if chemical formula and connectivity is the same
        """
        return self.inchi_key[:14] == other.inchi_key[:14]

    def fingerprint(
        self, radius: int, nbits: int = 2048, chiral: bool = False
    ) -> np.ndarray:
        """
        Returns the Morgan fingerprint of the molecule

        :param radius: the radius of the fingerprint
        :param nbits: the length of the fingerprint
        :param chiral: if True, include chirality information
        :return: the fingerprint
        """
        key = radius, nbits

        if key not in self._fingerprints:
            self.sanitize()
            bitvect = AllChem.GetMorganFingerprintAsBitVect(
                self.rd_mol, *key, useChirality=chiral
            )
            array = np.zeros((1,))
            DataStructs.ConvertToNumpyArray(bitvect, array)
            self._fingerprints[key] = array

        return self._fingerprints[key]

    def has_atom_mapping(self) -> bool:
        """
        Determines if a the molecule has atom mappings

        :return: True if at least one atom has a mapping
        """
        for atom in self.rd_mol.GetAtoms():
            if atom.GetAtomMapNum() > 0:
                return True
        return False

    def make_unique(self) -> "UniqueMolecule":
        """
        Returns an instance of the UniqueMolecule class that
        is representing the same molecule but is not hashable or comparable.

        :return: the unique molecule
        """
        return UniqueMolecule(rd_mol=self.rd_mol)

    def remove_atom_mapping(self, exceptions: Optional[Sequence[int]] = None) -> None:
        """
        Remove all mappings of the atoms and update the smiles

        :param exceptions: keep the listed atom mappings
        """
        exceptions = exceptions or []
        for atom in self.rd_mol.GetAtoms():
            if exceptions and atom.GetAtomMapNum() in exceptions:
                continue
            atom.SetAtomMapNum(0)
        self.smiles = Chem.MolToSmiles(self.rd_mol)
        self._clear_cache()

    def sanitize(self, raise_exception: bool = True) -> None:
        """
        Sanitizes the molecule if it has not been done before.

        :param raise_exception: if True will raise exception on failed sanitation
        :raises MoleculeException: if the molecule could not be sanitized
        """
        if self._is_sanitized:
            return

        try:
            AllChem.SanitizeMol(self.rd_mol)
        # pylint: disable=bare-except
        except:  # noqa, there could be many reasons why the molecule cannot be sanitized
            if raise_exception:
                raise MoleculeException(f"Unable to sanitize molecule ({self.smiles})")
            self.rd_mol = Chem.MolFromSmiles(self.smiles, sanitize=False)
            return

        self.smiles = Chem.MolToSmiles(self.rd_mol)
        self._clear_cache()
        self._is_sanitized = True

    def _clear_cache(self):
        self._inchi = None
        self._inchi_key = None
        self._fingerprints = {}
        self._atom_mappings = {}
        self._reverse_atom_mappings = {}


class TreeMolecule(Molecule):
    """
    A special molecule that keeps a reference to a parent molecule.

    If the class is instantiated without specifying the `transform` argument,
    it is computed by increasing the value of the `parent.transform` variable.

    If no parent is provided the atoms with atom mapping number are tracked
    and inherited to children.

    :ivar mapped_mol: the tracked molecule with atom mappings
    :ivar mapped_smiles: the SMILES of the tracked molecule with atom mappings
    :ivar original_smiles: the SMILES as passed when instantiating the class
    :ivar parent: parent molecule
    :ivar transform: a numerical number corresponding to the depth in the tree

    :param parent: a TreeMolecule object that is the parent
    :param transform: the transform value, defaults to None
    :param rd_mol: a RDKit mol object to encapsulate, defaults to None
    :param smiles: a SMILES to convert to a molecule object, defaults to None
    :param sanitize: if True, the molecule will be immediately sanitized, defaults to False
    :param mapping_update_callback: if given will call this method before setting up the `mapped_smiles`
    :raises MoleculeException: if neither rd_mol or smiles is given, or if the molecule could not be sanitized
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        parent: Optional["TreeMolecule"],
        transform: Optional[int] = None,
        rd_mol: Optional[RdMol] = None,
        smiles: Optional[str] = None,
        sanitize: bool = False,
        mapping_update_callback: Optional[Callable[["TreeMolecule"], None]] = None,
    ) -> None:
        super().__init__(rd_mol=rd_mol, smiles=smiles, sanitize=sanitize)
        self.parent = parent
        if transform is None and parent and parent.transform is not None:
            self.transform: int = parent.transform + 1
        else:
            self.transform = transform or 0

        self.original_smiles = smiles
        self.mapped_mol = Chem.Mol(self.rd_mol)
        self._atom_bonds: List[Tuple[int, int]] = []
        if not self.parent:
            self._set_atom_mappings()
        elif mapping_update_callback is not None:
            mapping_update_callback(self)

        AllChem.SanitizeMol(self.mapped_mol)
        self.mapped_smiles = Chem.MolToSmiles(self.mapped_mol)

        if self.parent:
            self.remove_atom_mapping()

    @property
    def mapping_to_index(self) -> Dict[int, int]:
        """Return a dictionary mapping to atom mappings to atom indices"""
        if not self._atom_mappings:
            self._atom_mappings = {
                atom.GetAtomMapNum(): atom.GetIdx()
                for atom in self.mapped_mol.GetAtoms()
                if atom.GetAtomMapNum()
            }
        return self._atom_mappings

    @property
    def mapped_atom_bonds(self) -> List[Tuple[int, int]]:
        """Return a list of atom bonds as tuples on the mapped atom indices"""
        bonds = []
        for bond in self.mapped_mol.GetBonds():
            bonds.append((bond.GetBeginAtom().GetIdx(), bond.GetEndAtom().GetIdx()))

        self._atom_bonds = [
            (self.index_to_mapping[atom_index1], self.index_to_mapping[atom_index2])
            for atom_index1, atom_index2 in bonds
        ]
        return self._atom_bonds

    def get_bonds_in_molecule(
        self, query_bonds: Sequence[Sequence[int]]
    ) -> Sequence[Sequence[int]]:
        """
        Get bonds (from a list of bonds) that are present in the molecule.
        :param bonds: List of bond (atom pairs)
        :return: A list of bonds
        """
        molecule_bonds = sort_bonds(self.mapped_atom_bonds)
        query_bonds = sort_bonds(query_bonds)
        bonds_in_mol = [bond for bond in query_bonds if bond in molecule_bonds]
        return bonds_in_mol

    def has_all_focussed_bonds(self, bonds: Sequence[Sequence[int]]) -> bool:
        """Checks that the focussed bonds exist in the target molecule's atom bonds.

        :param bonds: Focussed bonds.
        :param target_mol: The target molecule.

        :return: A boolean indicating if the input bonds exist in the target molecule.
        """
        bonds_in_mol = self.get_bonds_in_molecule(bonds)
        return len(bonds_in_mol) == len(bonds)

    def _set_atom_mappings(self) -> None:
        atom_mappings = [
            atom.GetAtomMapNum()
            for atom in self.mapped_mol.GetAtoms()
            if atom.GetAtomMapNum() != 0
        ]

        mapper = max(atom_mappings) + 1 if atom_mappings else 1
        self._atom_mappings = {}
        for atom_index, atom in enumerate(self.mapped_mol.GetAtoms()):
            if atom.GetAtomMapNum() == 0:
                atom.SetAtomMapNum(mapper)
                mapper += 1
            self._atom_mappings[atom.GetAtomMapNum()] = atom_index


class UniqueMolecule(Molecule):
    """
    A special molecule with the hash set to the id of the object.
    Therefore no two instances of this class will be comparable.

    :param rd_mol: a RDKit mol object to encapsulate, defaults to None
    :param smiles: a SMILES to convert to a molecule object, defaults to None
    :param sanitize: if True, the molecule will be immediately sanitized, defaults to False
    :raises MoleculeException: if neither rd_mol or smiles is given, or if the molecule could not be sanitized
    """

    def __init__(
        self,
        rd_mol: Optional[RdMol] = None,
        smiles: Optional[str] = None,
        sanitize: bool = False,
    ) -> None:
        super().__init__(rd_mol=rd_mol, smiles=smiles, sanitize=sanitize)

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, _) -> bool:
        return False


def none_molecule() -> UniqueMolecule:
    """Return an empty molecule"""
    return UniqueMolecule(rd_mol=Chem.MolFromSmiles(""))



""" Module containing the implementation of a reaction tree or route and factory classes to make such trees """

from __future__ import annotations

import abc
import hashlib
import json
import operator
from typing import TYPE_CHECKING

import networkx as nx
from networkx.algorithms.traversal.depth_first_search import dfs_tree

try:
    from route_distances.route_distances import route_distances_calculator
except ImportError:
    SUPPORT_DISTANCES = False
else:
    SUPPORT_DISTANCES = True

from aizynthfinder.chem import (
    FixedRetroReaction,
    Molecule,
    UniqueMolecule,
    none_molecule,
)
from aizynthfinder.utils.image import RouteImageFactory

if TYPE_CHECKING:
    from aizynthfinder.chem import RetroReaction
    from aizynthfinder.utils.type_utils import (
        Any,
        Dict,
        FrameColors,
        Iterable,
        Optional,
        PilImage,
        StrDict,
        Union,
    )


class ReactionTree:
    """
    Encapsulation of a bipartite reaction tree of a single route.
    The nodes consists of either FixedRetroReaction or UniqueMolecule objects.

    The reaction tree is initialized at instantiation and is not supposed to
    be updated.

    :ivar graph: the bipartite graph
    :ivar is_solved: if all of the leaf nodes are in stock
    :ivar root: the root of the tree
    :ivar created_at_iteration: iteration the reaction tree was created
    """

    def __init__(self) -> None:
        self.graph = nx.DiGraph()
        self.root = none_molecule()
        self.is_solved: bool = False
        self.created_at_iteration: Optional[int] = None

    @classmethod
    def from_dict(cls, tree_dict: StrDict) -> "ReactionTree":
        """
        Create a new ReactionTree by parsing a dictionary.

        This is supposed to be the opposite of ``to_dict``,
        but because that format loses information, the returned
        object is not a full copy as the stock will only contain
        the list of molecules marked as ``in_stock`` in the dictionary.

        The returned object should be sufficient to e.g. generate an image of the route.

        :param tree_dict: the dictionary representation
        :returns: the reaction tree
        """
        return ReactionTreeFromDict(tree_dict).tree

    @property
    def metadata(self) -> StrDict:
        """Return a dicitionary with route metadata"""
        return {
            "created_at_iteration": self.created_at_iteration,
            "is_solved": self.is_solved,
        }

    def depth(self, node: Union[UniqueMolecule, FixedRetroReaction]) -> int:
        """
        Return the depth of a node in the route

        :param node: the query node
        :return: the depth
        """
        return self.graph.nodes[node].get("depth", -1)

    def distance_to(self, other: "ReactionTree", content: str = "both") -> float:
        """
        Calculate the distance to another reaction tree

        This is a tree edit distance, with unit cost to
        insert and deleted nodes, and the Jaccard distance for substituting nodes

        :param other: the reaction tree to compare to
        :param content: determine what part of the tree to include in the calculation
        :return: the distance between the routes
        """
        if not SUPPORT_DISTANCES:
            raise ValueError(
                "Distance calculations are not supported by this installation."
                " Please install aizynthfinder with extras dependencies."
            )
        calculator = route_distances_calculator("ted", content=content)
        distances = calculator([self.to_dict(), other.to_dict()])
        return distances[0, 1]

    def hash_key(self) -> str:
        """
        Calculates a hash code for the tree using the sha224 hash function recursively

        :return: the hash key
        """
        return self._hash_func(self.root)

    def in_stock(self, node: Union[UniqueMolecule, FixedRetroReaction]) -> bool:
        """
        Return if a node in the route is in stock

        Note that is a property set on creation and as such is not updated.

        :param node: the query node
        :return: if the molecule is in stock
        """
        return self.graph.nodes[node].get("in_stock", False)

    def is_branched(self) -> bool:
        """
        Returns if the route is branched

        i.e. checks if the maximum depth is not equal to the number
        of reactions.
        """
        nsteps = len(list(self.reactions()))
        max_depth = max(self.depth(leaf) for leaf in self.leafs())
        return nsteps != max_depth // 2

    def leafs(self) -> Iterable[UniqueMolecule]:
        """
        Generates the molecules nodes of the reaction tree that has no predecessors,
        i.e. molecules that has not been broken down

        :yield: the next leaf molecule in the tree
        """
        for node in self.graph:
            if isinstance(node, UniqueMolecule) and not self.graph[node]:
                yield node

    def molecules(self) -> Iterable[UniqueMolecule]:
        """
        Generates the molecule nodes of the reaction tree

        :yield: the next molecule in the tree
        """
        for node in self.graph:
            if isinstance(node, UniqueMolecule):
                yield node

    def parent_molecule(self, mol: UniqueMolecule) -> UniqueMolecule:
        """Returns the parent molecule within the reaction tree.
        :param mol: the query node (molecule)
        :return: the parent molecule
        """
        if mol is self.root:
            raise ValueError("Root molecule does not have any parent node.")

        parent_reaction = list(self.graph.predecessors(mol))[0]
        parent_molecule = list(self.graph.predecessors(parent_reaction))[0]
        return parent_molecule

    def reactions(self) -> Iterable[FixedRetroReaction]:
        """
        Generates the reaction nodes of the reaction tree

        :yield: the next reaction in the tree
        """
        for node in self.graph:
            if not isinstance(node, Molecule):
                yield node

    def subtrees(self) -> Iterable[ReactionTree]:
        """
        Generates the subtrees of this reaction tree a
        subtree is a reaction treee starting at a molecule node that has children.

        :yield: the next subtree
        """

        def create_subtree(root_node):
            subtree = ReactionTree()
            subtree.root = root_node
            subtree.graph = dfs_tree(self.graph, root_node)
            for node in subtree.graph:
                prop = dict(self.graph.nodes[node])
                prop["depth"] -= self.graph.nodes[root_node].get("depth", 0)
                if "transform" in prop:
                    prop["transform"] -= self.graph.nodes[root_node].get("transform", 0)
                subtree.graph.nodes[node].update(prop)
            subtree.is_solved = all(subtree.in_stock(node) for node in subtree.leafs())
            return subtree

        for node in self.molecules():
            if node is not self.root and self.graph[node]:
                yield create_subtree(node)

    def to_dict(self, include_metadata=False) -> StrDict:
        """
        Returns the reaction tree as a dictionary in a pre-defined format.
        :param include_metadata: if True include metadata
        :return: the reaction tree
        """
        return self._build_dict(self.root, include_metadata=include_metadata)

    def to_image(
        self,
        in_stock_colors: Optional[FrameColors] = None,
        show_all: bool = True,
    ) -> PilImage:
        """
        Return a pictorial representation of the route

        :param in_stock_colors: the colors around molecules, defaults to {True: "green", False: "orange"}
        :param show_all: if True, also show nodes that are marked as hidden
        :return: the image of the route
        """
        factory = RouteImageFactory(
            self.to_dict(), in_stock_colors=in_stock_colors, show_all=show_all
        )
        return factory.image

    def to_json(self, include_metadata=False) -> str:
        """
        Returns the reaction tree as a JSON string in a pre-defined format.

        :return: the reaction tree
        """
        return json.dumps(
            self.to_dict(include_metadata=include_metadata), sort_keys=False, indent=2
        )

    def _build_dict(
        self,
        node: Union[UniqueMolecule, FixedRetroReaction],
        dict_: Optional[StrDict] = None,
        include_metadata=False,
    ) -> StrDict:
        if dict_ is None:
            dict_ = {}

        if node is self.root and include_metadata:
            dict_["route_metadata"] = self.metadata

        dict_["type"] = "mol" if isinstance(node, Molecule) else "reaction"
        dict_["hide"] = self.graph.nodes[node].get("hide", False)
        dict_["smiles"] = node.smiles
        if isinstance(node, UniqueMolecule):
            dict_["is_chemical"] = True
            dict_["in_stock"] = self.in_stock(node)
        elif isinstance(node, FixedRetroReaction):
            dict_["is_reaction"] = True
            dict_["metadata"] = dict(node.metadata)
        else:
            raise ValueError(
                f"This is an invalid reaction tree. Unknown node type {type(node)}"
            )

        dict_["children"] = []

        children = list(self.graph.successors(node))
        if isinstance(node, FixedRetroReaction):
            children.sort(key=operator.attrgetter("weight"))
        for child in children:
            child_dict = self._build_dict(child)
            dict_["children"].append(child_dict)

        if not dict_["children"]:
            del dict_["children"]
        return dict_

    def _hash_func(self, node: Union[FixedRetroReaction, UniqueMolecule]) -> str:
        if isinstance(node, UniqueMolecule):
            hash_ = hashlib.sha224(node.inchi_key.encode())
        else:
            hash_ = hashlib.sha224(node.hash_key().encode())
        child_hashes = sorted(
            self._hash_func(child) for child in self.graph.successors(node)
        )
        for child_hash in child_hashes:
            hash_.update(child_hash.encode())
        return hash_.hexdigest()


class ReactionTreeLoader(abc.ABC):
    """
    Base class for classes that creates a reaction tree object

    This class makes sure that node attributes are set after the
    graph is generated, and provides utility methods.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._unique_mols: Dict[int, UniqueMolecule] = {}
        self._unique_reactions: Dict[int, FixedRetroReaction] = {}
        self.tree = ReactionTree()
        self._load(*args, **kwargs)

        self.tree.is_solved = all(
            self.tree.in_stock(node) for node in self.tree.leafs()
        )

    def _add_node(
        self,
        node: Union[UniqueMolecule, FixedRetroReaction],
        depth: int = 0,
        transform: int = 0,
        in_stock: bool = False,
        hide: bool = False,
    ) -> None:
        attributes = {
            "hide": hide,
            "depth": depth,
        }
        if isinstance(node, Molecule):
            attributes.update({"transform": transform, "in_stock": in_stock})
            if not self.tree.root:
                self.tree.root = node
        self.tree.graph.add_node(node, **attributes)

    @abc.abstractmethod
    def _load(self, *args: Any, **kwargs: Any) -> None:
        pass

    def _unique_mol(self, molecule: Molecule) -> UniqueMolecule:
        id_ = id(molecule)
        if id_ not in self._unique_mols:
            self._unique_mols[id_] = molecule.make_unique()
        return self._unique_mols[id_]

    def _unique_reaction(self, reaction: RetroReaction) -> FixedRetroReaction:
        id_ = id(reaction)
        if id_ not in self._unique_reactions:
            metadata = dict(reaction.metadata)
            if ":" in reaction.mapped_reaction_smiles():
                metadata["mapped_reaction_smiles"] = reaction.mapped_reaction_smiles()
            self._unique_reactions[id_] = FixedRetroReaction(
                self._unique_mol(reaction.mol),
                smiles=reaction.smiles,
                metadata=metadata,
            )
        return self._unique_reactions[id_]


class ReactionTreeFromDict(ReactionTreeLoader):
    """Creates a reaction tree object from a dictionary"""

    def _load(self, tree_dict: StrDict) -> None:  # type: ignore
        if tree_dict.get("route_metadata"):
            self.tree.created_at_iteration = tree_dict["route_metadata"].get(
                "created_at_iteration"
            )
        self._parse_tree_dict(tree_dict)

    def _parse_tree_dict(self, tree_dict: StrDict, ncalls: int = 0) -> UniqueMolecule:
        product_node = UniqueMolecule(smiles=tree_dict["smiles"])
        self._add_node(
            product_node,
            depth=2 * ncalls,
            transform=ncalls,
            hide=tree_dict.get("hide", False),
            in_stock=tree_dict["in_stock"],
        )

        rxn_tree_dict = tree_dict.get("children", [])
        if not rxn_tree_dict:
            return product_node

        rxn_tree_dict = rxn_tree_dict[0]
        reaction_node = FixedRetroReaction(
            product_node,
            smiles=rxn_tree_dict["smiles"],
            metadata=rxn_tree_dict.get("metadata", {}),
        )
        self._add_node(
            reaction_node, depth=2 * ncalls + 1, hide=rxn_tree_dict.get("hide", False)
        )
        self.tree.graph.add_edge(product_node, reaction_node)

        reactant_nodes = []
        for reactant_tree in rxn_tree_dict.get("children", []):
            reactant_node = self._parse_tree_dict(reactant_tree, ncalls + 1)
            self.tree.graph.add_edge(reaction_node, reactant_node)
            reactant_nodes.append(reactant_node)
        reaction_node.reactants = (tuple(reactant_nodes),)

        return product_node


class ReactionTreeFromExpansion(ReactionTreeLoader):
    """
    Create a ReactionTree from a single reaction

    This is mainly intended as a convenience function for the expander interface
    """

    def _load(self, reaction: RetroReaction) -> None:  # type: ignore
        root = self._unique_mol(reaction.mol)
        self._add_node(root)

        rxn = self._unique_reaction(reaction)
        if hasattr(reaction, "smarts"):
            rxn.metadata["smarts"] = reaction.smarts  # type: ignore
        self._add_node(rxn)
        self.tree.graph.add_edge(root, rxn)

        reactant_nodes = []
        for reactant in reaction.reactants[0]:
            reactant_node = self._unique_mol(reactant)
            reactant_nodes.append(reactant_node)
            self._add_node(reactant_node)
            self.tree.graph.add_edge(rxn, reactant_node)
        rxn.reactants = (tuple(reactant_nodes),)



"""Module containing a GUI extension for plotting pareto fronts
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from IPython.display import display
from ipywidgets import Button, Checkbox, Dropdown, HBox, IntRangeSlider, Output, VBox
from aizynthfinder.analysis import (
    RouteCollection,
    RouteSelectionArguments,
    TreeAnalysis,
)
from aizynthfinder.interfaces.gui.utils import pareto_fronts_plot, route_display

if TYPE_CHECKING:
    from aizynthfinder.context.scoring.collection import ScorerCollection
    from aizynthfinder.interfaces.aizynthapp import AiZynthApp
    from aizynthfinder.search.andor_trees import AndOrSearchTreeBase
    from aizynthfinder.search.mcts import MctsSearchTree
    from aizynthfinder.utils.type_utils import Optional, StrDict, Union


class ParetoFrontsGUI:
    """GUI extension for plotting pareto fronts of routes.

    :param routes: the routes to plot
    :param scorers: the available scorers
    """

    def __init__(
        self,
        tree: Union[MctsSearchTree, AndOrSearchTreeBase],
        scorers: ScorerCollection,
    ) -> None:
        self._tree = tree
        self._scorers = scorers
        self._routes: Optional[RouteCollection] = None
        self._input: StrDict = dict()
        self._output: StrDict = dict()
        self._buttons: StrDict = dict()
        self._create_input()
        self._create_route_widgets()

    @classmethod
    def from_app(cls, app: AiZynthApp):
        """,
        Helper function to create a GUI from a GUI app interface

        :param app: the app to extract the routes from
        :return: the GUI object
        """
        if app.finder.tree is None:
            raise ValueError("Cannot initialize GUI from no tree")
        return ParetoFrontsGUI(app.finder.tree, app.finder.scorers)

    def _toggle_button(self, on_) -> None:
        for button in self._buttons.values():
            button.disabled = not on_

    def _create_input(self) -> None:
        self._input["scorer_1"] = Dropdown(
            options=self._scorers.names(),
            description="Scorer 1:",
            style={"description_width": "initial"},
        )
        self._input["scorer_2"] = Dropdown(
            options=self._scorers.names(),
            description="Scorer 2:",
            style={"description_width": "initial"},
        )
        self._input["select_routes"] = IntRangeSlider(
            value=[5, 25],
            min=1,
            max=50,
            step=1,
            description="Select #Routes:",
            disabled=False,
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format="d",
            style={"description_width": "initial"},
        )
        self._input["select_all_routes"] = Checkbox(
            value=False,
            description="Select All Routes",
            style={"description_width": "initial"},
            layout={"justify": "left"},
        )
        selection_box = HBox(
            [
                self._input["select_routes"],
                self._input["select_all_routes"],
            ]
        )
        scorers_box = HBox(
            [
                self._input["scorer_1"],
                self._input["scorer_2"],
            ]
        )
        display(
            VBox([selection_box, scorers_box], layout={"border": "1px solid silver"})
        )

    def _create_route_widgets(self) -> None:
        self._buttons["show_routes"] = Button(description="Show Routes")
        self._buttons["show_routes"].on_click(self._on_display_routes_button_clicked)
        self._input["route"] = Dropdown(
            options=[],
            description="Routes: ",
        )
        self._input["route"].observe(self._on_change_route_option)
        display(
            HBox(
                [
                    self._buttons["show_routes"],
                    self._input["route"],
                ]
            )
        )

        self._output["pareto_fronts"] = Output(
            layout={"border": "1px solid silver", "width": "99%", "overflow": "auto"}
        )
        display(self._output["pareto_fronts"])

        self._output["routes"] = Output(
            layout={"border": "1px solid silver", "width": "99%", "overflow": "auto"}
        )
        display(self._output["routes"])

    def _on_display_routes_button_clicked(self, _) -> None:
        self._toggle_button(False)

        objectives = [self._input["scorer_1"].value, self._input["scorer_2"].value]
        mo_scorers = [self._scorers[name] for name in objectives]
        analysis = TreeAnalysis(self._tree, scorer=mo_scorers)

        config_selection = RouteSelectionArguments(
            nmin=self._input["select_routes"].value[0],
            nmax=self._input["select_routes"].value[1],
            return_all=self._input["select_all_routes"].value,
        )
        self._routes = RouteCollection.from_analysis(analysis, config_selection)
        self._routes.compute_scores(*self._scorers.objects())
        self._routes.make_images()

        self._input["route"].options = [
            f"Option {i}" for i, _ in enumerate(self._routes, 1)  # type: ignore
        ]

        with self._output["pareto_fronts"]:
            pareto_fronts_plot(self._routes)

        self._show_route(0)
        self._toggle_button(True)

    def _on_change_route_option(self, change) -> None:
        if change["name"] != "index":
            return
        self._show_route(self._input["route"].index)

    def _show_route(self, index) -> None:
        if self._routes is None:
            return
        route_display(index, self._routes, self._output["routes"])

    def _reset_widgets(self):
        self._output["pareto_fronts"].clear_output()
        self._output["routes"].clear_output()
        self._input["route"].options = []



""" Sub-package containing MCTS routines
"""
from aizynthfinder.search.mcts.node import MctsNode
from aizynthfinder.search.mcts.search import MctsSearchTree
from aizynthfinder.search.mcts.state import MctsState



""" Module containing classes that interfaces different stocks query classes
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pandas as pd

try:
    import molbloom
except ImportError:
    HAS_MOLBLOOM = False
else:
    HAS_MOLBLOOM = True

from aizynthfinder.chem import Molecule
from aizynthfinder.utils.exceptions import StockException
from aizynthfinder.utils.mongo import get_mongo_client

if TYPE_CHECKING:
    from pymongo.collection import Collection as MongoCollection
    from pymongo.database import Database as MongoDatabase

    from aizynthfinder.utils.type_utils import Optional, Set, StrDict


class StockQueryMixin:
    """
    Mixin class for all query classes, providing a default interface
    to some methods that might not be possible to implement for each
    query class.
    """

    def __len__(self) -> int:
        return 0

    def __contains__(self, mol: Molecule) -> bool:
        return False

    def amount(self, mol: Molecule) -> float:
        """
        Returns the maximum amount of the molecule in stock

        :param mol: the query molecule
        :raises StockException: if the amount cannot be computed
        :return: the amount
        """
        raise StockException("Cannot compute amount")

    def availability_string(self, mol: Molecule) -> str:
        """
        Returns the sources of the molecule

        :param mol: the query molecule
        :raises StockException: if the string cannot be computed
        :return: a comma-separated list of sources
        """
        raise StockException("Cannot provide availability")

    def cached_search(self, mol: Molecule) -> bool:
        """
        Finds the entries of the molecule in the stock and cache them
        if necessary.

        :param mol: the query molecule
        :return: if the molecule is in stock
        """
        return mol in self

    def clear_cache(self) -> None:
        """Clear the internal search cache if available"""

    def price(self, mol: Molecule) -> float:
        """
        Returns the minimum price of the molecule in stock

        :param mol: the query molecule
        :raises StockException: if the price cannot be computed
        :rtype: float
        """
        raise StockException("Cannot compute price")


class InMemoryInchiKeyQuery(StockQueryMixin):
    """
    A stock query class that is based on an in-memory list
    of pre-computed inchi-keys.

    This list can be instantiated from
        * A Pandas dataframe in HDF5 or CSV format
        * A text file with an inchi key on each row

    The dataframe must have a column with InChIkeys that by default is "inchi_key".
    The HDF5 file must have a dataset called "table".

    If the source is a dataframe, then optionally it can contain prices and this
    columns can be specified with the "price_column" argument.

    :parameter path: the path to the file with inchi-keys
    :parameter inchi_key_col: the name of the column of the InChI keys
    :paramater price_col: the name of the column with the optional prices
    """

    def __init__(
        self,
        path: str,
        inchi_key_col: str = "inchi_key",
        price_col: Optional[str] = None,
    ) -> None:
        ext = os.path.splitext(path)[1]
        if ext not in [".h5", ".hdf5", ".csv"]:
            with open(path, "r") as fileobj:
                inchis = fileobj.read().splitlines()
            self._stock_inchikeys = set(inchis)
            self._price_dict: StrDict = {}
            return

        if ext in [".h5", ".hdf5"]:
            stock_df: pd.DataFrame = pd.read_hdf(path, key="table")
        else:
            stock_df = pd.read_csv(
                path,
                usecols=[inchi_key_col, price_col] if price_col else [inchi_key_col],
            )
        inchis = stock_df[inchi_key_col].values
        self._stock_inchikeys = set(inchis)

        if price_col is None:
            self._price_dict = {}
            return

        if len(stock_df) != len(self._stock_inchikeys):
            raise StockException(
                "InChI keys in stock df are expected to be unique, currently they are not"
            )
        if stock_df[price_col].isnull().sum() != 0:
            raise StockException(
                "null values found in the price column, please drop/impute them first"
            )
        if stock_df[price_col].min() < 0:
            raise StockException(
                f"expected non-negative prices, the min in the current file {path} is {stock_df[price_col].min()} "
            )
        self._price_dict = dict(zip(stock_df[inchi_key_col], stock_df[price_col]))

    def __contains__(self, mol: Molecule) -> bool:
        return mol.inchi_key in self._stock_inchikeys

    def __len__(self) -> int:
        return len(self._stock_inchikeys)

    @property
    def stock_inchikeys(self) -> Set[str]:
        """Return the InChiKeys in this stock"""
        return self._stock_inchikeys

    def price(self, mol: Molecule) -> float:
        if not self._price_dict:
            raise StockException(
                "no prices created, check the path type and if price column is supplied"
            )
        if mol in self:
            return self._price_dict[mol.inchi_key]
        raise StockException(f"no price info available for {mol.smiles}")


class MongoDbInchiKeyQuery(StockQueryMixin):
    """
    A stock query class that is looking up inchi keys in a Mongo database.

    The documents in the database collection should have at least 2 fields:
        * inchi_key: the inchi key of the molecule
        * source: the original source of the molecule

    :ivar client: the Mongo client
    :ivar database: the database instance
    :ivar molecules: the collection of documents

    :parameter host: the database host, defaults to None
    :parameter database: the database name, defaults to "stock_db"
    :parameter collection: the database collection, defaults to "molecules"
    """

    def __init__(
        self,
        host: Optional[str] = None,
        database: str = "stock_db",
        collection: str = "molecules",
    ) -> None:
        self.client = get_mongo_client(
            host or os.environ.get("MONGODB_HOST") or "localhost"
        )
        if self.client is None:
            raise ImportError(
                "Cannot use this stock query class because it seems like pymongo is not installed. "
                "Please install aizynthfinder with extras dependencies."
            )
        self.database: MongoDatabase = self.client[database]
        self.molecules: MongoCollection = self.database[collection]
        self._len: Optional[int] = None

    def __contains__(self, mol: Molecule) -> bool:
        return self.molecules.count_documents({"inchi_key": mol.inchi_key}) > 0

    def __len__(self) -> int:
        if self._len is None:
            self._len = self.molecules.count_documents({})
        return self._len

    def __str__(self) -> str:
        return "'MongoDB stock'"

    def availability_string(self, mol: Molecule) -> str:
        sources = [
            item["source"] for item in self.molecules.find({"inchi_key": mol.inchi_key})
        ]
        return ",".join(sources)


class MolbloomFilterQuery(StockQueryMixin):
    """
    A stock query class that is based on an a molbloom filter
    for SMILES strings or InChI keys

    :parameter path: the path to the saved bloom filter
    :parameter smiles_based: if True will use SMILES for lookup instead of InChI keys
    """

    def __init__(self, path: str, smiles_based: bool = False) -> None:
        if not HAS_MOLBLOOM:
            raise ImportError(
                "Cannot use this stock query class because it seems like molbloom is not installed. "
                "Please install aizynthfinder with extras dependencies."
            )
        self._filter = molbloom.BloomFilter(path)
        self._smiles_based = smiles_based

    def __contains__(self, mol: Molecule) -> bool:
        if self._smiles_based:
            return mol.smiles in self._filter
        return mol.inchi_key in self._filter


STOCK_QUERY_ALIAS = {
    "inchiset": "InMemoryInchiKeyQuery",
    "mongodb": "MongoDbInchiKeyQuery",
    "bloom": "MolbloomFilterQuery",
}
